#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/html/htmldbg.cpp,v 1.1.1.1 1999/05/09 21:32:28 MJRoberts Exp $";
#endif

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  htmldbg.cpp - debugger helper class implementation
Function
  
Notes
  
Modified
  02/04/98 MJRoberts  - Creation
*/

#include <string.h>

/* include necessary TADS interfaces */
#include <dbg.h>
#include <err.h>
#include <lin.h>
#include <linf.h>
#include <dat.h>
#include <os.h>
#include <bif.h>
#include <obj.h>

#ifndef TADSHTML_H
#include "tadshtml.h"
#endif
#ifndef HTMLDBG_H
#include "htmldbg.h"
#endif
#ifndef HTMLSYS_H
#include "htmlsys.h"
#endif
#ifndef HTMLPRS_H
#include "htmlprs.h"
#endif
#ifndef HTMLFMT_H
#include "htmlfmt.h"
#endif
#ifndef HTMLTAGS_H
#include "htmltags.h"
#endif
#ifndef HTMLDISP_H
#include "htmldisp.h"
#endif
#ifndef HTMLDCFG_H
#include "htmldcfg.h"
#endif

/* the hidden output status variable */
extern "C" int dbghid;

CHtmlDebugHelper::CHtmlDebugHelper()
{
    /* there are no windows in the source window list yet */
    first_srcwin_ = last_srcwin_ = 0;

    /* no current or context line yet */
    cur_is_valid_ = FALSE;
    ctx_is_valid_ = FALSE;
    ctx_stackwin_disp_ = 0;

    /* no stack window yet */
    stack_win_ = 0;

    /* no debug log window yet */
    debuglog_win_ = 0;

    /* no history window yet */
    hist_win_ = 0;

    /* set default tab expansion size of 8 */
    tab_size_ = 8;

    /* no temporary breakpoint yet */
    tmp_bp_valid_ = FALSE;
}

CHtmlDebugHelper::~CHtmlDebugHelper()
{
    /* close any remaining source windows */
    while (first_srcwin_ != 0)
        first_srcwin_->win_->close_window(TRUE);

    /* delete any remaining special line links */
    while (first_line_ != 0)
    {
        CHtmlDbg_line_link *cur;
        
        /* remember this item */
        cur = first_line_;

        /* remove it from the list */
        first_line_ = first_line_->next_;
        if (first_line_ != 0)
            first_line_->prev_ = 0;

        /* delete this item */
        delete cur;
    }
}


/* 
 *   set execution state to GO 
 */
void CHtmlDebugHelper::set_exec_state_go(dbgcxdef *ctx)
{
    /* clear the stepping flags */
    ctx->dbgcxflg &= ~(DBGCXFSS + DBGCXFSO);

    /* remove the old current line pointer while exeucting */
    forget_current_line();
}

/* 
 *   set execution state to STEP OVER - steps without following calls 
 */
void CHtmlDebugHelper::set_exec_state_step_over(dbgcxdef *ctx)
{
    /* set the step and step-over-calls flags */
    ctx->dbgcxflg |= (DBGCXFSS + DBGCXFSO);

    /* keep stepping until we're back at the current depth again */
    ctx->dbgcxsof = ctx->dbgcxdep;

    /* remove the old current line pointer while exeucting */
    forget_current_line();
}

/*
 *   set execution state to STEP OUT - steps until the current function
 *   has returned 
 */
void CHtmlDebugHelper::set_exec_state_step_out(dbgcxdef *ctx)
{
    /* set the step and step-over-calls flags */
    ctx->dbgcxflg |= (DBGCXFSS + DBGCXFSO);

    /* keep stepping until we're back at the *enclosing* depth */
    ctx->dbgcxsof = (ctx->dbgcxdep >= 1 ? ctx->dbgcxdep - 1 : 0);

    /* remove the old current line pointer while exeucting */
    forget_current_line();
}

/* 
 *   set execution state to STEP INTO - steps and follows calls 
 */
void CHtmlDebugHelper::set_exec_state_step_into(dbgcxdef *ctx)
{
    /* set the stepping flag, but clear the step-over-calls flag */
    ctx->dbgcxflg |= DBGCXFSS;
    ctx->dbgcxflg &= ~DBGCXFSO;

    /* remove the old current line pointer while exeucting */
    forget_current_line();
}

/*
 *   Signal a QUIT condition 
 */
void CHtmlDebugHelper::signal_quit(dbgcxdef *ctx)
{
    /* signal the TADS error condition for quitting the game */
    errsig(ctx->dbgcxerr, ERR_RUNQUIT);
}

/*
 *   Signal a RESTART condition 
 */
void CHtmlDebugHelper::signal_restart(dbgcxdef *ctx)
{
    /* 
     *   clear the in-debugger flag - we're leaving the debugger via the
     *   signal, so we don't need to keep this flag set 
     */
    ctx->dbgcxflg &= ~DBGCXFIND;

    /* clear the debugger stack */
    ctx->dbgcxdep = 0;
    ctx->dbgcxfcn = 0;

    /* 
     *   go restart via the built-in restart() function, which takes care
     *   of resetting all the run-time context information 
     */
    bifres((bifcxdef *)ctx->dbgcxrun->runcxbcx, 0);
}

/*
 *   Signal an ABORT condition 
 */
void CHtmlDebugHelper::signal_abort(dbgcxdef *ctx)
{
    /* signal the TADS error condition for quitting the game */
    errsig(ctx->dbgcxerr, ERR_RUNABRT);
}

/*
 *   Move execution to the current selection in the given window.  Returns
 *   zero on success, non-zero on error.  
 */
int CHtmlDebugHelper::set_next_statement(dbgcxdef *ctx,
                                         unsigned int *exec_ofs,
                                         CHtmlSysWin *win)
{
    CHtmlDbg_win_link *win_link;
    unsigned long fpos;
    objnum line_objn;
    uint line_ofs;

    /* make sure it's valid */
    if (!is_in_same_fn(ctx, exec_ofs, win))
        return 1;

    /* find information on the current line in the window */
    if (srcofs_to_code(ctx, win, &fpos, &win_link, &line_objn, &line_ofs))
        return 2;

    /* move the execution point to the new offset */
    *exec_ofs = line_ofs;

    /* update the current line display */
    forget_current_line();
    set_line_status(win_link->source_id_, fpos, HTMLTDB_STAT_CURLINE);

    /* success */
    return 0;
}

/*
 *   Check if the current selection in the given window is within the same
 *   function or method as the current execution point.  Returns true if
 *   so, false if not.  This is a condition of setting the next statement;
 *   this function is provided so that the UI code can check to make sure
 *   that this condition is met when the user attempts to move the
 *   execution point, and generate an appropriate error message if not.  
 */
int CHtmlDebugHelper::is_in_same_fn(dbgcxdef *ctx,
                                    unsigned int *exec_ofs,
                                    CHtmlSysWin *win)
{
    CHtmlDbg_win_link *win_link;
    unsigned long fpos;
    objnum line_objn;
    uint line_ofs;
    uint start_ofs, end_ofs;
    dbgfdef *fr;

    /* get the current frame - if there isn't one, return failure */
    if (ctx->dbgcxfcn == 0)
        return FALSE;
    fr = &ctx->dbgcxfrm[ctx->dbgcxfcn - 1];

    /* find information on the current line in the window */
    if (srcofs_to_code(ctx, win, &fpos, &win_link, &line_objn, &line_ofs))
        return FALSE;

    /* 
     *   if this line is not in the same object as the current target
     *   object, we can disallow it 
     */
    if (line_objn == MCMONINV || line_objn != fr->dbgftarg)
        return FALSE;

    /* 
     *   if there's no property, it's a function, so it's valid as long as
     *   it's in the current object (which we've just determined that it
     *   is) 
     */
    if (fr->dbgfprop == (prpnum)0)
        return TRUE;

    /* find the limits of the property's data range */
    start_ofs = objgetp(ctx->dbgcxmem, line_objn, fr->dbgfprop, 0);
    end_ofs = objgetp_end(ctx->dbgcxmem, line_objn, fr->dbgfprop);

    /* if it's within the range, it's valid, otherwise it's not */
    return (line_ofs >= start_ofs && line_ofs < end_ofs);
}



/*
 *   get the current call trace status 
 */
int CHtmlDebugHelper::is_call_trace_active(dbgcxdef *ctx) const
{
    return (ctx->dbgcxflg & DBGCXFTRC) != 0;
}

/*
 *   turn call trace on or off 
 */
void CHtmlDebugHelper::set_call_trace_active(dbgcxdef *ctx, int flag)
{
    if (flag)
        ctx->dbgcxflg |= DBGCXFTRC;
    else
        ctx->dbgcxflg &= ~DBGCXFTRC;
}

/*
 *   clear the call trace log 
 */
void CHtmlDebugHelper::clear_call_trace_log(dbgcxdef *ctx)
{
    /* forget everything in the buffer */
    ctx->dbgcxhstf = 0;

    /* make sure we reflect this in the display window, if open */
    update_hist_window(ctx);
}

/*
 *   get a pointer to the history log buffer 
 */
const textchar_t *CHtmlDebugHelper::get_call_trace_buf(dbgcxdef *ctx) const
{
    return ctx->dbgcxhstp;
}

/*
 *   get the size of the data accumulated in the history log 
 */
unsigned long CHtmlDebugHelper::get_call_trace_len(dbgcxdef *ctx) const
{
    return ctx->dbgcxhstf;
}

/*
 *   Get the object and p-code offset for the current selection in the
 *   given window.  Returns zero on success, non-zero on failure.  On
 *   return, *fpos will be changed to reflect the actual source file
 *   location closest to the input fpos value that contains a line start.
 *   *win_link is filled in with a pointer to the window link structure
 *   for the given window.  
 */
int CHtmlDebugHelper::srcofs_to_code(dbgcxdef *ctx, CHtmlSysWin *win,
                                     unsigned long *fpos,
                                     CHtmlDbg_win_link **win_link,
                                     objnum *line_objn, uint *line_ofs)
{
    lindef *lin;
    uchar posbuf[4];
    CHtmlDbg_line_link *line_link;

    /* find information the current line in the window */
    *fpos = find_line_info(win, win_link, &line_link);

    /* if we didn't find the window, return failure */
    if (*win_link == 0)
        return 1;

    /* get the line source */
    lin = CHtmlDbg_win_link::find_lindef(ctx, (*win_link)->source_id_);
    if (lin == 0)
        return 2;

    /* find the source line at this location */
    oswp4(posbuf, *fpos);
    linfind(lin, (char *)posbuf, line_objn, line_ofs);

    /* 
     *   adjust the file position to the actual vlaue -- this may be
     *   different than the text the user was pointing at, since not every
     *   line contains executable code 
     */
    *fpos = osrp4(posbuf);

    /* make sur ewe have a valid location */
    if (*line_objn == MCMONINV)
        return 3;

    /* success */
    return 0;
}

/*
 *   Set or clear a breakpoint at the current selection in the given
 *   window 
 */
void CHtmlDebugHelper::toggle_breakpoint(dbgcxdef *ctx,
                                         CHtmlSysWin *win)
{
    CHtmlDbg_win_link *win_link;
    unsigned long fpos;
    objnum line_objn;
    uint line_ofs;
    int bpnum;
    int did_set;
    char bpname[HTMLDBG_MAXBPNAME];
    lindef *lin;

    /* find information on the current line in the window */
    if (srcofs_to_code(ctx, win, &fpos, &win_link, &line_objn, &line_ofs))
        return;
    
    /* find the line source */
    lin = CHtmlDbg_win_link::find_lindef(ctx, win_link->source_id_);
    if (lin == 0)
        return;

    /*
     *   Build the name of the breakpoint for display and configuration
     *   saving purposes.  The name consists of the name of the line
     *   source and the source offset.  
     */
    sprintf(bpname, "#%lu %s", fpos, ((linfdef *)lin)->linfnam);

    /* set the breakpoint */
    if (dbgbpat(ctx, line_objn, MCMONINV, line_ofs, &bpnum,
                bpname, TRUE, 0, &did_set))
        return;

    /* update the special status of the line */
    change_line_status(win_link->source_id_, fpos,
                       HTMLTDB_STAT_BP, did_set);

    /* 
     *   if we cleared the breakpoint, make sure we clear any
     *   breakpoint-related flags on the line
     */
    if (!did_set)
        change_line_status(win_link->source_id_, fpos,
                           (HTMLTDB_STAT_BP_DIS | HTMLTDB_STAT_BP_COND),
                           FALSE);
}

/*
 *   Set a temporary breakpoint at the current selection in the given
 *   window.  If there's already a breakpoint at this location, we won't
 *   add a new breakpoint, and we'll return false.  Otherwise, we'll set a
 *   new breakpoint and return true.
 *   
 *   The purpose of a temporary breakpoint is to allow "run to
 *   cursor"-type operations, where the debugger must secretly set its own
 *   internal breakpoint to facilitate the operation but where the user
 *   doesn't intend to add a new breakpoint to the persistent breakpoint
 *   list.  Because the breakpoint is temporary, we won't bother giving it
 *   a meaningful name for the display list, and we also won't show the
 *   breakpoint in the line status displayed in the source window; this
 *   breakpoint should thus be removed the next time the debugger is
 *   entered.
 *   
 *   Only one temporary breakpoint is allowed at a time.  If there's
 *   already a temporary breakpoint set, we'll remove the old one before
 *   adding the new one.  
 */
int CHtmlDebugHelper::set_temp_bp(dbgcxdef *ctx, CHtmlSysWin *win)
{
    CHtmlDbg_win_link *win_link;
    unsigned long fpos;
    objnum line_objn;
    uint line_ofs;
    int did_set;

    /* if we already have a temporary breakpoint, clear it */
    clear_temp_bp(ctx);

    /* find information on the current line in the window */
    if (srcofs_to_code(ctx, win, &fpos, &win_link, &line_objn, &line_ofs))
        return FALSE;

    /* if there's already a breakpoint there, there's nothing to do */
    if (dbgisbp(ctx, line_objn, MCMONINV, line_ofs, &tmp_bpnum_))
        return FALSE;

    /* set the breakpoint */
    if (dbgbpat(ctx, line_objn, MCMONINV, line_ofs, &tmp_bpnum_,
                "Temp BP", FALSE, 0, &did_set) || !did_set)
        return FALSE;

    /* 
     *   we've set the breakpoint -- note that we now have a temporary
     *   breakpoint, and return success indication 
     */
    tmp_bp_valid_ = TRUE;
    return TRUE;
}

/*
 *   Clear the temporary breakpoint, if there is one.  If no temporary
 *   breakpoint is currently set, this will have no effect. 
 */
void CHtmlDebugHelper::clear_temp_bp(dbgcxdef *ctx)
{
    /* if we don't have a temporary breakpoint, there's nothing to do */
    if (!tmp_bp_valid_)
        return;

    /* clear the breakpoint */
    dbgbpdel(ctx, tmp_bpnum_);

    /* forget the breakpoint */
    tmp_bp_valid_ = FALSE;
}

/*
 *   Add a new global breakpoint.  Returns a non-zero error code if the
 *   condition is invalid or any other error occurs; returns zero on
 *   success, and fills in *bpnum with the identifier of the new
 *   breakpoint.  
 */
int CHtmlDebugHelper::set_global_breakpoint(dbgcxdef *ctx,
                                            const textchar_t *cond,
                                            int *bpnum)
{
    /* 
     *   Add the breakpoint.  Use the name "g" for all global breakpoints;
     *   since the breakpoint is based entirely on the condition string,
     *   we don't need to remember anything about it but that it has a
     *   condition string. 
     */
    return dbgbpat(ctx, MCMONINV, MCMONINV, 0, bpnum, "g", TRUE,
                   (textchar_t *)cond, 0);
}

/*
 *   Enable/disable a breakpoint at the current selection point in the
 *   given window.  Reverses current enabling state.  
 */
void CHtmlDebugHelper::toggle_bp_disable(dbgcxdef *ctx,
                                         CHtmlSysWin *win)
{
    CHtmlDbg_win_link *win_link;
    unsigned long fpos;
    objnum line_objn;
    uint line_ofs;
    int bpnum;
    int disabled;

    /* find information on the current line in the window */
    if (srcofs_to_code(ctx, win, &fpos, &win_link, &line_objn, &line_ofs))
        return;

    /* 
     *   ask the debugger if there's a breakpoint there - if not, there's
     *   nothing to do 
     */
    if (!dbgisbp(ctx, line_objn, MCMONINV, line_ofs, &bpnum))
        return;

    /* toggle the enabled/disabled state of the breakpoint */
    disabled = dbgisbpena(ctx, bpnum);
    dbgbpdis(ctx, bpnum, disabled);

    /* update the special status of the line */
    change_line_status(win_link->source_id_, fpos,
                       HTMLTDB_STAT_BP_DIS, disabled);
}


/*
 *   Delete a breakpoint given a breakpoint number 
 */
int CHtmlDebugHelper::delete_breakpoint(dbgcxdef *ctx, int bpnum)
{
    int err;
    int has_line;
    int source_id;
    unsigned long fpos;

    /* identify the source line for the breakpoint */
    has_line = get_bp_line_info(ctx, bpnum, &source_id, &fpos);

    /* tell TADS to delete the breakpoint */
    if ((err = dbgbpdel(ctx, bpnum)) != 0)
        return err;

    /* update the line status */
    if (has_line)
        change_line_status(source_id, fpos, HTMLTDB_STAT_BP, FALSE);

    /* return the result */
    return err;
}

/*
 *   Delete all breakpoints 
 */
void CHtmlDebugHelper::delete_all_bps(dbgcxdef *ctx)
{
    int i;
    
    /* go through all breakpoints, and delete each one */
    for (i = 0 ; i < DBGBPMAX ; ++i)
        delete_breakpoint(ctx, i);
}

/*
 *   Enable or disable a breakpoint, given a breakpoint number 
 */
int CHtmlDebugHelper::enable_breakpoint(dbgcxdef *ctx, int bpnum, int enable)
{
    int err;
    int has_line;
    int source_id;
    unsigned long fpos;

    /* identify the source line for the breakpoint */
    has_line = get_bp_line_info(ctx, bpnum, &source_id, &fpos);

    /* tell TADS to enable or disable the breakpoint */
    if ((err = dbgbpdis(ctx, bpnum, !enable)) != 0)
        return err;

    /* update the line status */
    if (has_line)
        change_line_status(source_id, fpos, HTMLTDB_STAT_BP_DIS, !enable);

    /* return the result */
    return err;
}

/*
 *   Determine if a given breakpoint is associated with code.  Returns
 *   true if so, false if not. 
 */
int CHtmlDebugHelper::is_bp_at_line(struct dbgcxdef *ctx, int bpnum)
{
    int source_id;
    unsigned long fpos;

    /* 
     *   Look up the line information; return the success indication, but
     *   throw away the details, since the caller only cares whether the
     *   breakpoint is at code, not exactly where it is.  
     */
    return get_bp_line_info(ctx, bpnum, &source_id, &fpos);
}

/*
 *   Go to the code associated with a given breakpoint.  Simply shows the
 *   code; doesn't establish the code's stack context.  Returns the window
 *   associated with the code; returns null if the breakpoint isn't
 *   associated with code.  
 */
CHtmlSysWin *CHtmlDebugHelper::show_bp_code(struct dbgcxdef *ctx, int bpnum,
                                            CHtmlDebugSysIfc_win *syswinifc)
{
    lindef *lin;
    int source_id;
    unsigned long fpos;
    CHtmlDbg_win_link *link;
    CHtmlDisp *disp;
    
    /* 
     *   get the line information associated with the breakpoint; if there
     *   isn't any line information available, return failure 
     */
    if (!get_bp_line_info(ctx, bpnum, &source_id, &fpos))
        return 0;

    /* get the line source */
    for (lin = ctx->dbgcxlin ; lin != 0 ; lin = lin->linnxt)
    {
        if (lin->linid == source_id)
            break;
    }

    /* if we didn't find the line source, give up */
    if (lin == 0)
        return 0;

    /* load or find the source window for this line source */
    link = load_source_file(ctx, syswinifc, (linfdef *)lin, TRUE);
    if (link == 0)
        return 0;

    /* find the line at this position */
    disp = link->formatter_->find_by_debugsrcpos(fpos);

    /* scroll the source winodw to show this location */
    if (disp != 0)
    {
        CHtmlRect pos;

        /* get the position of this line, and scroll it into view */
        pos = disp->get_pos();
        link->win_->scroll_to_doc_coords(&pos);

        /* put the selection at teh start of the line */
        link->formatter_->set_sel_range(disp->get_text_ofs(),
                                        disp->get_text_ofs());
    }

    /* return the window we found */
    return link->win_;
}

/* callback structure for enumerating breakpoints for display */
struct enum_bp_disp_cb_ctx_t
{
    /* client callback to invoke with descriptive information */
    void (*cbfunc_)(void *, int, const textchar_t *, int);
    void *cbctx_;
};

/*
 *   Enumerate breakpoints in a displayable format 
 */
void CHtmlDebugHelper::
   enum_breakpoints(dbgcxdef *dbgctx,
                    void (*cbfunc)(void *cbctx, int bpnum,
                                   const textchar_t *desc, int disabled),
                    void *cbctx)
{
    enum_bp_disp_cb_ctx_t ourctx;
    
    /* set up our callback context */
    ourctx.cbfunc_ = cbfunc;
    ourctx.cbctx_ = cbctx;
    
    /* enumerate the breakpoints */
    dbgbpenum(dbgctx, &disp_bp_enumcb, &ourctx);
}

/*
 *   breakpoint enumerator for generating displayable breakpoints 
 */
void CHtmlDebugHelper::disp_bp_enumcb(void *ctx0, int bpnum,
                                      const char *privatedesc,
                                      const char *cond, int disabled)
{
    enum_bp_disp_cb_ctx_t *ctx = (enum_bp_disp_cb_ctx_t *)ctx0;
    textchar_t desc[HTMLDBG_MAXBPDESC];

    /* format the breakpoint description for display */
    format_bp_for_display(desc, sizeof(desc), privatedesc, cond, disabled);

    /* call the client callback */
    (*ctx->cbfunc_)(ctx->cbctx_, bpnum, desc, disabled);
}

/*
 *   Get a breakpoint's description in a format suitable for display.
 *   Returns zero on success, non-zero on failure.  
 */
int CHtmlDebugHelper::get_bp_desc(dbgcxdef *ctx, int bpnum,
                                  textchar_t *buf, size_t buflen)
{
    char desc[HTMLDBG_MAXBPNAME];
    char cond[HTMLDBG_MAXCOND];
    int disabled;

    /* get the information on this breakpoint */
    if (dbgbpgetinfo(ctx, bpnum, desc, sizeof(desc), cond, sizeof(cond)))
        return 1;

    /* check if it's disabled */
    disabled = !dbgisbpena(ctx, bpnum);

    /* format the text into the caller's buffer */
    format_bp_for_display(buf, buflen, desc, cond, disabled);

    /* success */
    return 0;
}

/*
 *   Determine if a breakpoint is enabled or disabled.  Returns true if
 *   the breakpoint is enabled, false if it's disabled.  
 */
int CHtmlDebugHelper::is_bp_enabled(dbgcxdef *ctx, int bpnum)
{
    return dbgisbpena(ctx, bpnum);
}

/*
 *   Get a breakpoint's condition.  Returns zero on success, non-zero on
 *   failure.  If this is a valid breakpoint without a condition, we'll
 *   return success and an empty condbuf string.  
 */
int CHtmlDebugHelper::get_bp_cond(dbgcxdef *ctx, int bpnum,
                                  textchar_t *condbuf, size_t condbuflen)
{
    /* get the information on this breakpoint into the caller's buffer */
    return dbgbpgetinfo(ctx, bpnum, 0, 0, condbuf, condbuflen);
}

/*
 *   Change a breakpoint's condition 
 */
int CHtmlDebugHelper::change_bp_cond(dbgcxdef *ctx, int bpnum,
                                     const textchar_t *cond)
{
    /* tell the debugger to do the work */
    return dbgbpsetcond(ctx, bpnum, (textchar_t *)cond);
}

/*
 *   Format a breakpoint description from our internal format to a format
 *   suitable for display to the user.  
 */
void CHtmlDebugHelper::format_bp_for_display(textchar_t *buf,
                                             size_t /*buflen*/,
                                             const textchar_t *privatedesc,
                                             const textchar_t *cond,
                                             int disabled)
{
    const textchar_t *p;
    
    /* check the first character to determine the breakpoint type */
    switch(privatedesc[0])
    {
    case '#':
        /* it's a source line - parse the file location information */
        for (p = privatedesc + 1 ; *p != ' ' ; ++p) ;

        /* format it as "at '<filename>'.<seekoffset> when <condition>" */
        sprintf(buf, "at '%s'.%.*s%s%s%s", p + 1,
                p - privatedesc - 1, privatedesc + 1,
                cond != 0 && cond[0] != '\0' ? " when " : "",
                cond != 0 ? cond : "",
                disabled ? " [disabled]" : "");

        /* done */
        break;

    case 'g':
        /* it's a global breakpoint - format as "when <condition>" */
        sprintf(buf, "when %s%s", cond,
                disabled ? " [disabled]" : "");
        break;

    default:
        /* invalid format - ignore it */
        break;
    }
}

/* 
 *   context structure for breakpoint enumeration callback for saving
 *   breakpoints to the configuration file 
 */
struct save_bp_enumcb_ctx_t
{
    CHtmlDebugConfig *config_;
    int count_;
};

/*
 *   Save breakpoint configuration 
 */
void CHtmlDebugHelper::save_bp_config(dbgcxdef *dbgctx,
                                      CHtmlDebugConfig *config)
{
    save_bp_enumcb_ctx_t cbctx;

    /* delete any existing configuration information for the breakpoints */
    config->clear_strlist("breakpoints", "desc");
    config->clear_strlist("breakpoints", "status");
    config->clear_strlist("breakpoints", "cond");

    /* enumerate the breakpoints, and let the callback do the work */
    cbctx.config_ = config;
    cbctx.count_ = 0;
    dbgbpenum(dbgctx, &save_bp_enumcb, &cbctx);
}

/*
 *   Breakpoint enumeration callback for saving breakpoint status to
 *   configuration 
 */
void CHtmlDebugHelper::save_bp_enumcb(void *ctx0, int /*bpnum*/,
                                      const char *desc,
                                      const char *cond, int disabled)
{
    save_bp_enumcb_ctx_t *ctx = (save_bp_enumcb_ctx_t *)ctx0;

    /* save this breakpoint's description */
    ctx->config_->setval("breakpoints", "desc", ctx->count_, desc);

    /* save it's enabled/disabled status */
    ctx->config_->setval("breakpoints", "status", ctx->count_,
                         disabled ? "d" : "e");

    /* if there's a condition, save it */
    ctx->config_->setval("breakpoints", "cond", ctx->count_,
                         cond != 0 ? cond : "");

    /* count it */
    ctx->count_++;
}

/*
 *   Load breakpoint configuration 
 */
void CHtmlDebugHelper::load_bp_config(dbgcxdef *dbgctx,
                                      CHtmlDebugConfig *config)
{
    int cnt;
    int i;
    
    /* run through the breakpoints in the configuration information */
    cnt = config->get_strlist_cnt("breakpoints", "desc");
    for (i = 0 ; i < cnt ; ++i)
    {
        textchar_t desc[HTMLDBG_MAXBPNAME];
        textchar_t stat[5];
        textchar_t cond[HTMLDBG_MAXCOND];
        int enabled;
        
        /* get this breakpoint's description */
        if (config->getval("breakpoints", "desc", i, desc, sizeof(desc)))
            continue;

        /* get its status */
        config->getval("breakpoints", "status", i, stat, sizeof(stat));
        enabled = (stat[0] == 'e');

        /* get the condition */
        if (config->getval("breakpoints", "cond", i, cond, sizeof(cond)))
            cond[0] = '\0';

        /* see what kind of breakpoint we have */
        switch(desc[0])
        {
        case '#':
            /* break at a source line */
            {
                unsigned long fpos;
                textchar_t *fname;
                lindef *lin;
                uchar posbuf[4];
                objnum line_objn;
                uint line_ofs;
                int bpnum;
                int did_set;
                unsigned int stat;

                /* get the source offset */
                fpos = (unsigned long)get_atol(desc + 1);

                /* find the filename portion */
                for (fname = desc + 1 ; *fname != '\0' && *fname != ' ' ;
                     ++fname) ;
                ++fname;

                /* find the line source with the given name */
                for (lin = dbgctx->dbgcxlin ; lin != 0 ; lin = lin->linnxt)
                {
                    /* if this is the one, we can stop looking */
                    if (get_strcmp(fname, ((linfdef *)lin)->linfnam) == 0)
                        break;
                }

                /* if we didn't find the line source, ignore the breakpoint */
                if (lin == 0)
                    break;

                /* find the source line at this location */
                oswp4(posbuf, fpos);
                linfind(lin, (char *)posbuf, &line_objn, &line_ofs);

                /* 
                 *   find the source line containing the breakpoint --
                 *   this may have changed, since the file may have been
                 *   edited since the last session 
                 */
                fpos = osrp4(posbuf);

                /* set the breakpoint */
                stat = HTMLTDB_STAT_BP;
                if (dbgbpat(dbgctx, line_objn, MCMONINV, line_ofs, &bpnum,
                            desc, TRUE, cond, &did_set))
                    break;

                /* disable the breakpoint if appropriate */
                if (did_set && !enabled)
                {
                    dbgbpdis(dbgctx, bpnum, TRUE);
                    stat |= HTMLTDB_STAT_BP_DIS;
                }

                /* toggle the special status of the line */
                change_line_status(lin->linid, fpos, stat, did_set);
            }
            break;

        case 'g':
            /* global breakpoint - set based on condition alone */
            {
                int bpnum;

                /* set the breakpoint */
                if (set_global_breakpoint(dbgctx, cond, &bpnum) == 0)
                {
                    /* disable the breakpoint if appropriate */
                    if (!enabled)
                        dbgbpdis(dbgctx, bpnum, TRUE);
                }
            }
            break;

        default:
            /* invalid format - ignore it */
            break;
        }
    }
}

/*
 *   callback structure for eval_expr_cb 
 */
struct eval_expr_cb_ctx_t
{
    /* debugger context */
    struct dbgcxdef *dbgctx;
    
    /* pointer to start of buffer and original length */
    textchar_t *buf;
    size_t buflen;

    /* pointer to the current position in the output buffer */
    textchar_t *outp;

    /* number of characters remaining in output buffer */
    size_t rem;

    /* logged error number */
    int err;

    /* flag indicating whether we want to keep logged error messages */
    int log_errors;
};

/*
 *   Get the text of an error message.  This routine doesn't require an
 *   error stack, so it's suitable for retrieving the text of error
 *   messages returned from the debugger API. 
 */
void CHtmlDebugHelper::get_error_text(struct dbgcxdef *ctx, int err,
                                      textchar_t *buf, size_t buflen)
{
    /* get the text of the message */
    errmsg(ctx->dbgcxerr, buf, buflen, err);
}

/*
 *   Get the text of an error message, formatting the message with
 *   arguments from the error stack.  This should only be used when a
 *   valid error stack is present, such as when entering the debugger due
 *   to a run-time error.  
 */
void CHtmlDebugHelper::format_error(struct dbgcxdef *ctx, int err,
                                    textchar_t *buf, size_t buflen)
{
    errcxdef *errctx = ctx->dbgcxerr;
    char msgtxt[256];

    /* get the text of the message */
    errmsg(errctx, msgtxt, sizeof(msgtxt), err);

    /* format it with arguments from the error stack */
    errfmt(buf, buflen, msgtxt,
           errctx->errcxptr->erraac, errctx->errcxptr->erraav);
}

/*
 *   Evaluate an expression 
 */
int CHtmlDebugHelper::eval_expr(struct dbgcxdef *ctx,
                                textchar_t *buf, size_t buflen,
                                const textchar_t *expr, int *is_lval,
                                int *is_openable,
                                void (*aggcb)(void *, const char *, int,
                                              const char *),
                                void *aggctx, int speculative)
{
    int err;
    int level;
    eval_expr_cb_ctx_t cbctx;
    void (*old_errcxlog)(void *, char *, int, int, erradef *);
    void  *old_errcxlgc;
    dattyp dat;

    /* 
     *   Set up the callback context to point to the output buffer.
     *   Reserve one character of the buffer for null termination. 
     */
    cbctx.dbgctx = ctx;
    cbctx.outp = cbctx.buf = buf;
    cbctx.rem = cbctx.buflen = buflen - 1;
    cbctx.err = 0;
    cbctx.log_errors = TRUE;

    /* if we're not in the debugger, we can't evaluate the expression */
    if (!(ctx->dbgcxflg & DBGCXFIND))
    {
        /* put the message in the buffer */
        err = ERR_DBGINACT;
        errmsg(ctx->dbgcxerr, buf, buflen, err);

        /* return the error */
        return err;
    }

    /* 
     *   get the frame level - note that our stack level counter has level
     *   0 as the innermost (current) level, which is the opposite of the
     *   way that dbgeval() counts levels, so we need to invert our
     *   counter to get the one to use in the dbgeval() call 
     */
    if (ctx_stackwin_disp_ != 0)
        level = ctx->dbgcxfcn - ctx_stackwin_disp_->get_debugsrcpos() - 1;
    else
        level = ctx->dbgcxfcn - 1;

    /* make sure we have a valid level, in case we have no stack frames */
    if (level < 0)
    {
        err = ERR_INACTFR;
        errmsg(ctx->dbgcxerr, buf, buflen, err);
        return err;
    }

    /*
     *   While we're evaluating, set up to receive error logging messages
     *   -- we'll write them to our output buffer and note the errors.
     *   Certain expression evaluation errors result in logged (rather
     *   than signalled) errors, so we need to trap any logged errors.  
     */
    old_errcxlog = ctx->dbgcxerr->errcxlog;
    old_errcxlgc = ctx->dbgcxerr->errcxlgc;
    ctx->dbgcxerr->errcxlog = eval_expr_errlog;
    ctx->dbgcxerr->errcxlgc = &cbctx;

    /* evaluate the expression */
    err = dbgevalext(ctx, (char *)expr, eval_expr_cb, &cbctx, level, FALSE,
                     &dat, aggcb, aggctx, speculative);

    /*
     *   determine whether the type is openable, and inform the caller if
     *   desired 
     */
    if (is_openable != 0)
    {
        switch(dat)
        {
        case DAT_LIST:
        case DAT_OBJECT:
            /* these types are openable */
            *is_openable = TRUE;
            break;

        default:
            /* other types are not openable */
            *is_openable = FALSE;
            break;
        }
    }

    /* null-terminate the value buffer */
    *(cbctx.outp) = '\0';

    /* check for error */
    if (err != 0)
    {
        char msg[256];
        
        /* an error occurred - put its text into the buffer */
        errmsg(ctx->dbgcxerr, msg, sizeof(msg), err);
        errfmt(buf, buflen, msg, 0, 0);
    }
    else if (cbctx.err != 0)
    {
        /* a logged error occurred; it's already in the buffer */
        err = cbctx.err;
    }
    else
    {
        /*
         *   If the caller wants to know, check the expression to
         *   determine if it's an lvalue.  To do this, try parsing "(expr)
         *   := 0"; if the expression is an lvalue, it will accept this,
         *   otherwise a parsing error will occur.  
         */
        if (is_lval != 0)
        {
            char asibuf[1024];
            objnum objn;

            /* 
             *   this parse will be speculative, so don't keep any error
             *   messages that get logged 
             */
            cbctx.log_errors = FALSE;

            /* build the assignment and evaluate it */
            sprintf(asibuf, "(%s):=0", expr);
            if (dbgcompile(ctx, asibuf, &ctx->dbgcxfrm[level], &objn,
                           speculative) != 0
                || cbctx.err != 0)
            {
                /* it's not an lvalue, so it can't be assigned into */
                *is_lval = FALSE;
            }
            else
            {
                /* it can be used as an lvalue */
                *is_lval = TRUE;
                
                /* 
                 *   free the compiled code - we didn't need it except for
                 *   parsing it 
                 */
                mcmfre(ctx->dbgcxmem, objn);
            }
        }
    }

    /* if an error occurred, it's definitely not an lvalue */
    if (err != 0 && is_lval != 0)
        *is_lval = FALSE;

    /* restore error context */
    ctx->dbgcxerr->errcxlog = old_errcxlog ;
    ctx->dbgcxerr->errcxlgc = old_errcxlgc;

    /* return the result */
    return err;
}

/*
 *   Display callback for eval_expr - simply adds text to the output
 *   buffer 
 */
void CHtmlDebugHelper::eval_expr_cb(void *ctx0, char *txt, int len)
{
    eval_expr_cb_ctx_t *ctx = (eval_expr_cb_ctx_t *)ctx0;

    /* if an error has been logged, don't record any more output */
    if (ctx->err != 0)
        return;

    /* if this is a single newline, omit it */
    if (len == 1 && *txt == '\n')
        return;

    /* limit the copy to the amount of room we have */
    if ((size_t)len > ctx->rem)
        len = (int)ctx->rem;

    /* copy the characters, turning newlines into spaces */
    while (len > 0)
    {
        *(ctx->outp++) = (*txt == '\n' ? ' ' : *txt);
        ++txt;
        --(ctx->rem);
        --len;
    }
}

/*
 *   Error logging callback for expression evaluator.  We'll put the error
 *   into our buffer and note the error. 
 */
void CHtmlDebugHelper::eval_expr_errlog(void *ctx0, char * /*fac*/, int err,
                                        int argc, erradef *argv)
{
    char msg[256];
    eval_expr_cb_ctx_t *ctx = (eval_expr_cb_ctx_t *)ctx0;

    /* keep the message only if we're currently logging errors */
    if (ctx->log_errors)
    {
        /* format the message into the buffer at the start of the buffer */
        errmsg(ctx->dbgctx->dbgcxerr, msg, sizeof(msg), err);
        errfmt(ctx->buf, ctx->buflen, msg, argc, argv);
        
        /* position the write pointer after the message */
        ctx->outp = ctx->buf + get_strlen(ctx->buf);
        ctx->rem = ctx->buflen - get_strlen(ctx->buf);
    }

    /* record the error */
    ctx->err = err;
}

/*
 *   Enumerate local variables in the current stack context.  
 */
void CHtmlDebugHelper::
   enum_locals(dbgcxdef *ctx,
               void (*cbfunc)(void *, const textchar_t *, size_t),
               void *cbctx)
{
    int level;
    
    /* if we're not in the debugger, there's no local variable context */
    if (!(ctx->dbgcxflg & DBGCXFIND))
        return;

    /* get the frame level */
    if (ctx_stackwin_disp_ != 0)
        level = ctx->dbgcxfcn - ctx_stackwin_disp_->get_debugsrcpos() - 1;
    else
        level = ctx->dbgcxfcn - 1;

    /* ask the debugger to do the enumeration */
    dbgenumlcl(ctx, level, cbfunc, cbctx);
}

/*
 *   Forget the current line.  This should be called just prior to
 *   resuming execution to ensure that the current line doesn't hang
 *   around while the game is executing for an extended period.  
 */
void CHtmlDebugHelper::forget_current_line()
{
    /* forget the current line if we have one */
    if (cur_is_valid_)
    {
        /* drop its current-line status flag */
        clear_line_status(cur_source_id_, cur_seekpos_, HTMLTDB_STAT_CURLINE);

        /* forget we have a current line */
        cur_is_valid_ = FALSE;
    }

    /* if we have a context line, drop it as well */
    forget_ctx_line();
}

/*
 *   forget the context line 
 */
void CHtmlDebugHelper::forget_ctx_line()
{
    if (ctx_is_valid_)
    {
        /* drop the context-line status flag from the source line */
        clear_line_status(ctx_source_id_, ctx_seekpos_, HTMLTDB_STAT_CTXLINE);

        /* forget about it */
        ctx_is_valid_ = FALSE;
    }

    /* clear the status flag in the stack window */
    if (ctx_stackwin_disp_ != 0 && stack_win_ != 0)
    {
        /* remove the flag */
        ctx_stackwin_disp_->set_dbg_status(stack_win_->win_, 0);
        
        /* forget the display item */
        ctx_stackwin_disp_ = 0;
    }
}

/*
 *   Update the debugger for a change in the execution status.  This
 *   should be called whenever we re-enter the debugger after executing
 *   game code.  We'll show the current line, reset the evaluation
 *   context, and update the stack window.  
 */
CHtmlSysWin *CHtmlDebugHelper::
   update_for_entry(dbgcxdef *ctx, CHtmlDebugSysIfc_win *syswinifc)
{
    CHtmlSysWin *win;

    /* load the current source line */
    win = load_current_source(ctx, syswinifc);

    /* update the stack window */
    update_stack_window(ctx, TRUE);

    /* update the call history window, if we're keeping call history */
    if ((ctx->dbgcxflg & DBGCXFTRC) != 0)
        update_hist_window(ctx);

    /* return the current source line window */
    return win;
}

/*
 *   Load the source file at a particular stack level.  Opens the window,
 *   scrolls to the current line, and sets the appropriate line status at
 *   that line.  
 */
CHtmlSysWin *CHtmlDebugHelper::
   load_source_at_level(dbgcxdef *ctx, CHtmlDebugSysIfc_win *syswinifc,
                        int level)
{
    uchar infobuf[128];
    lindef *lin;
    unsigned long fpos;
    CHtmlDbg_win_link *link;

    /* if we previously had a current line showing on-screen, remove it */
    if (level == 0)
    {
        /* forget the current and context lines */
        forget_current_line();
    }
    else
    {
        /* forget only the old context line */
        forget_ctx_line();
    }

    /* get information on the current line */
    if (dbglgetlvl(ctx, infobuf, level))
        return 0;

    /* search for a line source with the given ID */
    for (lin = ctx->dbgcxlin ; lin != 0 ; lin = lin->linnxt)
    {
        /* if this is the one, stop looking */
        if (lin->linid == infobuf[0])
            break;
    }

    /* if we didn't find a line source with the current line, give up */
    if (lin == 0)
        return 0;

    /* load or find the source window for this line source */
    link = load_source_file(ctx, syswinifc, (linfdef *)lin, TRUE);

    /* get the seek location of the current line */
    fpos = osrp4(infobuf + 1);

    /* go to the current line if we got a window */
    if (link != 0 && link->win_ != 0)
    {
        CHtmlDisp *disp;
        CHtmlRect pos;
        
        /* find the line at this position */
        disp = link->formatter_->find_by_debugsrcpos(fpos);

        /* scroll the source window to show this location */
        if (disp != 0)
        {
            /* bring the selection into view */
            pos = disp->get_pos();
            link->win_->scroll_to_doc_coords(&pos);
            
            /* put the selection at the start of this line */
            link->formatter_->set_sel_range(disp->get_text_ofs(),
                                            disp->get_text_ofs());
        }
    }

    /* set the current or the context line, as appropriate */
    if (level == 0)
    {
        /* set the new current line */
        cur_is_valid_ = TRUE;
        cur_source_id_ = lin->linid;
        cur_seekpos_ = fpos;

        /* update the status */
        set_line_status(cur_source_id_, cur_seekpos_, HTMLTDB_STAT_CURLINE);
    }
    else
    {
        /* set the new context line */
        ctx_is_valid_ = TRUE;
        ctx_source_id_ = lin->linid;
        ctx_seekpos_ = fpos;

        /* update the status */
        set_line_status(ctx_source_id_, ctx_seekpos_, HTMLTDB_STAT_CTXLINE);
    }

    /* return the source window we found */
    return link->win_;
}

/*
 *   Enumerate open windows 
 */
void CHtmlDebugHelper::
   enum_source_windows(void (*func)(void *ctx, int idx,
                                    class CHtmlSysWin *win,
                                    int line_source_id,
                                    HtmlDbg_wintype_t win_type),
                       void *cbctx)
{
    CHtmlDbg_win_link *link;
    int idx;

    /* go through the list of windows */
    for (link = first_srcwin_, idx = 0 ; link != 0 ;
         link = link->next_, ++idx)
    {
        /* invoke the callback for this window */
        (*func)(cbctx, idx, link->win_, link->source_id_, link->win_type_);
    }
}

/*
 *   Get a source window by index 
 */
CHtmlSysWin *CHtmlDebugHelper::get_source_window_by_index(int idx) const
{
    CHtmlDbg_win_link *link;

    /* go through the list of windows */
    for (link = first_srcwin_ ; link != 0 && idx != 0 ;
         link = link->next_, --idx) ;

    /* return the window we found */
    return (link == 0 ? 0 : link->win_);
}

/*
 *   Open a window on a line source, given the line source ID 
 */
CHtmlSysWin *CHtmlDebugHelper::
   open_line_source(dbgcxdef *ctx, CHtmlDebugSysIfc_win *syswinifc,
                    int line_source_id)
{
    lindef *lin;
    
    /* scan the line sources to find the one with the given ID */
    for (lin = ctx->dbgcxlin ; lin != 0 ; lin = lin->linnxt)
    {
        /* if the ID matches, open this one */
        if (line_source_id == lin->linid)
        {
            CHtmlDbg_win_link *link;
            
            /* open it */
            link = load_source_file(ctx, syswinifc, (linfdef *)lin, TRUE);

            /* return the window */
            return (link == 0 ? 0 : link->win_);
        }
    }

    /* didn't find it */
    return 0;
}

/*
 *   Load an arbitrary text file into a window.
 */
CHtmlSysWin *CHtmlDebugHelper::
   open_text_file(dbgcxdef *ctx, CHtmlDebugSysIfc_win *syswinifc,
                  const char *fname)
{
    struct
    {
        linfdef linf;
        char fname[256];
    } linf;
    osfildef *fp;
    CHtmlDbg_win_link *link;
    lindef *lin;

    /*
     *   Scan the source file line sources to see if we're actually
     *   opening one of the game's source files.  If so, just open the
     *   line source directly. 
     */
    for (lin = ctx->dbgcxlin ; lin != 0 ; lin = lin->linnxt)
    {
        /* see if this name matches that of the line source */
        if (syswinifc->dbgsys_cmp_fnames(fname, ((linfdef *)lin)->linfnam))
        {
            /* it's a match - go open the line source directly */
            link = load_source_file(ctx, syswinifc, (linfdef *)lin, TRUE);

            /* return the line source's window */
            return (link == 0 ? 0 : link->win_);
        }
    }

    /* open the file */
    fp = osfoprs(fname, OSFTTEXT);
    if (fp == 0)
        return 0;

    /* set up a line source to read the file */
    linfini2(ctx->dbgcxmem, &linf.linf, (char *)fname, get_strlen(fname), fp);

    /* 
     *   this is not one of the game's source files, so give it an invalid
     *   line source ID 
     */
    linf.linf.linflin.linid = -1;

    /* load it */
    link = load_source_file(ctx, syswinifc, &linf.linf, FALSE);

    /* done with the source file - close it */
    osfcls(fp);

    /* return the window */
    return (link == 0 ? 0 : link->win_);
}

/*
 *   Load a source file into a window.  If the source file is already
 *   loaded, we'll simply return the existing window.  
 */
CHtmlDbg_win_link *CHtmlDebugHelper::
   load_source_file(dbgcxdef *ctx, CHtmlDebugSysIfc_win *syswinifc,
                    linfdef *linf, int need_activate)
{
    CHtmlDbg_win_link *link;
    
    /*
     *   Search for the file in our open file list.  If we find it, simply
     *   return the existing window.  
     */
    for (link = first_srcwin_ ; link != 0 ; link = link->next_)
    {
        /* if the filename matches the one we want, stop looking */
        if (syswinifc->dbgsys_cmp_fnames(linf->linfnam, link->fname_.get()))
            return link;
    }

    /*
     *   We didn't find the file in our open file list, so create a new
     *   window and load the file into the new window. 
     */
    link = create_source_win(syswinifc, linf);
    load_file_into_win(ctx, link, linf, need_activate);

    /* return the link */
    return link;
}

/*
 *   Load a source file into a given window 
 */
void CHtmlDebugHelper::load_file_into_win(dbgcxdef *ctx,
                                          CHtmlDbg_win_link *win_link,
                                          linfdef *linf, int need_activate)
{
    lindef *lin;
    uchar posbuf[4];
    
    /* get the base class line source */
    lin = &linf->linflin;
    
    /* activate the line source to open the file, if it's not already open */
    if (need_activate)
        linactiv(lin);

    /* if we don't have a valid file, we can't proceed */
    if (linf->linffp == 0)
        return;

    /* seek to the start fo the line source */
    oswp4(posbuf, 0);
    linseek(lin, posbuf);

    /* make sure there's nothing on the page */
    win_link->parser_->clear_page();

    /* 
     *   read lines from the line source, and add each line to the
     *   formatter list for the window 
     */
    for (;;)
    {
        textchar_t buf[1024];
        CHtmlTagText *txt_tag;
        textchar_t *p;
        unsigned long fpos;
        uchar posbuf[4];
        
        /* note the source file position of this line start */
        lintell(lin, posbuf);
        fpos = osrp4(posbuf);

        /* get the next line */
        if (!lingets(lin, (uchar *)buf, sizeof(buf)))
            break;

        /* remove any newline characters at the end of the line */
        for (p = buf + get_strlen(buf) ; p > buf ; )
        {
            /* move to the previous character */
            --p;
            
            /* if it's not a \r or \n, we're done */
            if (*p != '\n' && *p != '\r')
                break;

            /* remove it and continue scanning */
            *p = '\0';
        }

        /* add a text tag for the line source */
        txt_tag =
            new CHtmlTagTextDebugsrc(fpos, tab_size_,
                                     win_link->parser_->get_text_array(),
                                     buf, get_strlen(buf));
        win_link->parser_->append_tag(txt_tag);
    }

    /* do the initial formatting of the window */
    win_link->win_->do_formatting(TRUE, TRUE, TRUE);

    /* update the scrollbars */
    win_link->win_->fmt_adjust_vscroll();

    /* update special-line indicators for this window */
    update_lines_on_screen(win_link);

    /* done with the file - deactive the line source if we activate it */
    if (need_activate)
        lindisact(lin);
}

/*
 *   Go through our special-line records and update the display so that a
 *   window's settings reflect the current special-line status flags.
 *   Note that we only have to do this for line sources with valid (i.e.,
 *   other than -1) id's, since these are the only sources that can have
 *   special line status flags.  
 */
void CHtmlDebugHelper::update_lines_on_screen(CHtmlDbg_win_link *win_link)
{
    CHtmlDbg_line_link *line_link;

    if (win_link->source_id_ != -1)
    {
        /* look through our list of special lines */
        for (line_link = first_line_ ; line_link != 0 ;
             line_link = line_link->next_)
        {
            /* if this line is from the source we just loaded, update it */
            if (line_link->source_id_ == win_link->source_id_)
            {
                /* update this line on-screen */
                update_line_on_screen(win_link, line_link);
            }
        }
    }
}


/*
 *   Create a new window for displaying a source file
 */
CHtmlDbg_win_link *CHtmlDebugHelper::create_source_win(
    CHtmlDebugSysIfc_win *syswinifc, linfdef *linf)
{
    return create_win(syswinifc, linf, linf->linfnam,
                      HTMLDBG_WINTYPE_SRCFILE);
}

/*
 *   Set the tab expansion size to use in source windows.  If this changes
 *   while windows are open, we'll need to reload files into all open
 *   windows, since we translate tabs while loading files. 
 */
void CHtmlDebugHelper::set_tab_size(struct dbgcxdef *ctx,
                                    class CHtmlDebugSysIfc_win *syswinifc,
                                    int n)
{
    if (tab_size_ != n && tab_size_ != 0)
    {
        /* remember the new tab size */
        tab_size_ = n;
        
        /* reload all source windows */
        reload_source_windows(ctx, syswinifc);
    }
}

/*
 *   Reformat all source windows 
 */
void CHtmlDebugHelper::
   reformat_source_windows(dbgcxdef *ctx, CHtmlDebugSysIfc_win *syswinifc)
{
    CHtmlDbg_win_link *link;

    /* go through the list of windows, and reformat each */
    for (link = first_srcwin_ ; link != 0 ; link = link->next_)
        reformat_source_window(ctx, syswinifc, link);
}

/*
 *   Reformat a source file 
 */
void CHtmlDebugHelper::
   reformat_source_window(dbgcxdef *ctx, CHtmlDebugSysIfc_win *syswinifc,
                          CHtmlDbg_win_link *link)
{
    CHtmlFontDesc font_desc;

    switch(link->win_type_)
    {
    case HTMLDBG_WINTYPE_SRCFILE:
    case HTMLDBG_WINTYPE_HISTORY:
        /* reset the window */
        link->formatter_->start_at_top(TRUE);
        
        /* get the font, in case it's changed */
        font_desc.fixed_pitch = TRUE;
        font_desc.htmlsize = 3;
        syswinifc->dbgsys_set_srcfont(&font_desc);
        link->font_ = link->win_->get_font(&font_desc);
        link->formatter_->set_font(link->font_);
        
        /* do a complete reformat of this window */
        link->win_->do_formatting(TRUE, TRUE, TRUE);
        link->win_->fmt_adjust_vscroll();
        
        /* make sure we've updated line indicators */
        update_lines_on_screen(link);
        break;

    case HTMLDBG_WINTYPE_STACK:
        /* 
         *   for the stack window, just rebuild it from scratch - it's
         *   easier than duplicating the code that figures out where all
         *   the special indicators go 
         */
        update_stack_window(ctx, FALSE);
        break;
    }
}

/*
 *   Reload all source windows 
 */
void CHtmlDebugHelper::
   reload_source_windows(struct dbgcxdef *ctx,
                         CHtmlDebugSysIfc_win *syswinifc)
{
    CHtmlDbg_win_link *link;

    /* go through the list of windows */
    for (link = first_srcwin_ ; link != 0 ; link = link->next_)
    {
        /* if it's a source file, reload it */
        if (link->win_type_ == HTMLDBG_WINTYPE_SRCFILE)
        {
            struct
            {
                linfdef linf;
                char fname[256];
            } linf;
            osfildef *fp = 0;
            
            /* open the file */
            fp = osfoprs(link->fname_.get(), OSFTTEXT);
            if (fp == 0)
                continue;
            
            /* set up a new line source */
            linfini2(ctx->dbgcxmem, &linf.linf,
                     (char *)link->fname_.get(),
                     get_strlen(link->fname_.get()), fp);
            
            /* load it */
            load_file_into_win(ctx, link, &linf.linf, FALSE);
        }
            
        /* reformat it */
        reformat_source_window(ctx, syswinifc, link);
    }
}

/*
 *   get the stack window 
 */
CHtmlSysWin *CHtmlDebugHelper::get_stack_win() const
{
    return (stack_win_ != 0 ? stack_win_->win_ : 0);
}

/*
 *   get the history window 
 */
CHtmlSysWin *CHtmlDebugHelper::get_hist_win() const
{
    return (hist_win_ != 0 ? hist_win_->win_ : 0);
}

/*
 *   get the debug log window 
 */
CHtmlSysWin *CHtmlDebugHelper::get_debuglog_win() const
{
    return (debuglog_win_ != 0 ? debuglog_win_->win_ : 0);
}

/*
 *   Create a window to use for stack display 
 */
CHtmlSysWin *CHtmlDebugHelper::
   create_stack_win(CHtmlDebugSysIfc_win *syswinifc)
{
    /* create a new stack window if we don't already have one */
    if (stack_win_ == 0)
        stack_win_ = create_win(syswinifc, 0, "Stack",
                                HTMLDBG_WINTYPE_STACK);

    /* return the system window */
    return stack_win_->win_;
}

/*
 *   Create the debug log window 
 */
CHtmlSysWin *CHtmlDebugHelper::
   create_debuglog_win(CHtmlDebugSysIfc_win *syswinifc)
{
    /* create a new debug log window if we don't have one yet */
    if (debuglog_win_ == 0)
        debuglog_win_ = create_win(syswinifc, 0, "Debug Log",
                                   HTMLDBG_WINTYPE_DEBUGLOG);

    /* return the system window */
    return debuglog_win_->win_;
}

/*
 *   Create a window to use for history display 
 */
CHtmlSysWin *CHtmlDebugHelper::
   create_hist_win(CHtmlDebugSysIfc_win *syswinifc)
{
    /* create a new stack window if we don't already have one */
    if (hist_win_ == 0)
        hist_win_ = create_win(syswinifc, 0, "Call History",
                               HTMLDBG_WINTYPE_HISTORY);

    /* return the system window */
    return hist_win_->win_;
}


/*
 *   Create a debugger window.  linf can be null (this will be the case
 *   for system windows, such as the stack window, not tied to a line
 *   source).  
 */
CHtmlDbg_win_link *CHtmlDebugHelper::
   create_win(CHtmlDebugSysIfc_win *syswinifc, linfdef *linf,
              const textchar_t *win_title, HtmlDbg_wintype_t win_type)
{
    CHtmlParser *parser;
    CHtmlFormatter *formatter;
    CHtmlSysWin *win;
    CHtmlFontDesc font_desc;
    CHtmlSysFont *font;

    /* create a parser for the window */
    parser = new CHtmlParser(TRUE);

    /* create a formatter for the window */
    formatter = new CHtmlFormatterMain(parser);

    /* create the system window using the system interface */
    win = syswinifc->dbgsys_create_win(parser, formatter, win_title,
                                       win_type);

    /* set up the formatter with the default monospaced font */
    font_desc.fixed_pitch = TRUE;
    font_desc.htmlsize = 3;
    syswinifc->dbgsys_set_srcfont(&font_desc);
    font = win->get_font(&font_desc);
    formatter->set_font(font);

    /* 
     *   add the window to our list of source windows, and return the new
     *   link structure 
     */
    return add_src_win(win, parser, formatter, linf, win_title,
                       font, win_type);
}

/*
 *   Add a window to the source window list 
 */
CHtmlDbg_win_link *CHtmlDebugHelper::add_src_win(
    CHtmlSysWin *win, CHtmlParser *parser, CHtmlFormatter *formatter,
    linfdef *linf, const textchar_t *win_title, CHtmlSysFont *font,
    HtmlDbg_wintype_t win_type)
{
    CHtmlDbg_win_link *link;
    
    /* create the new link */
    link = new CHtmlDbg_win_link(win, parser, formatter,
                                 linf, win_title, font, win_type);

    /* link it in at the end of our list of source windows */
    if (last_srcwin_ != 0)
        last_srcwin_->next_ = link;
    else
        first_srcwin_ = link;
    link->prev_ = last_srcwin_;
    last_srcwin_ = link;

    /* return the new link */
    return link;
}

/*
 *   Delete a window link from the source window list.  Removes the link
 *   from the list and deletes the link object.  
 */
void CHtmlDebugHelper::del_src_win(CHtmlDbg_win_link *link)
{
    /* unlink it from its previous item */
    if (link->prev_ != 0)
        link->prev_->next_ = link->next_;
    else
        first_srcwin_ = link->next_;

    /* unlink it from its next item */
    if (link->next_ != 0)
        link->next_->prev_ = link->prev_;
    else
        last_srcwin_ = link->prev_;

    /* delete the link */
    delete link;
}

/*
 *   Receive notification that a source window is being closed 
 */
void CHtmlDebugHelper::on_close_srcwin(CHtmlSysWin *win)
{
    CHtmlDbg_win_link *link;
    
    /* find the window in our list */
    for (link = first_srcwin_ ; link != 0 ; link = link->next_)
    {
        /* if this is the one, delete it */
        if (link->win_ == win)
        {
            /* this is it - delete the formatter and parser */
            link->formatter_->release_parser();
            delete link->parser_;
            delete link->formatter_;

            /* if this is the stack window, forget that as well */
            if (link == stack_win_)
                stack_win_ = 0;

            /* if it's the debug log window, forget it */
            if (link == debuglog_win_)
                debuglog_win_ = 0;

            /* if it's the history window, forget it, too */
            if (link == hist_win_)
                hist_win_ = 0;

            /* now delete the link itself */
            del_src_win(link);

            /* we're done */
            return;
        }
    }

    /* we didn't find it; ignore the request */
}

/*
 *   Get the line and window link records for the current selection in a
 *   given window.  Returns the file seek position of the start of the
 *   line containing the start of the selection.  
 */
unsigned long CHtmlDebugHelper::
   find_line_info(CHtmlSysWin *win, CHtmlDbg_win_link **win_linkp,
                  CHtmlDbg_line_link **line_linkp)
{
    CHtmlDbg_win_link *wlink;
    
    /* presume we won't find anything */
    *win_linkp = 0;
    *line_linkp = 0;

    /* find the window in our list of source windows */
    for (wlink = first_srcwin_ ; wlink != 0 ; wlink = wlink->next_)
    {
        /* if this is the one, find the current selection in it */
        if (wlink->win_ == win)
        {
            unsigned long sel_start, sel_end;
            unsigned long fpos;
            CHtmlDispTextDebugsrc *disp;

            /* get the start of the current selection */
            wlink->formatter_->get_sel_range(&sel_start, &sel_end);

            /* find the display item at the start of the selection */
            disp = (CHtmlDispTextDebugsrc *)
                   wlink->formatter_->find_by_txtofs(sel_start, TRUE, FALSE);
            if (disp == 0)
                return 0;

            /* get the file position of the display item */
            fpos = disp->get_debugsrcpos();

            /* tell the caller about the window */
            *win_linkp = wlink;

            /* find the line record */
            *line_linkp = find_line_link(wlink->source_id_, fpos);

            /* return the file position */
            return fpos;
        }
    }

    /* didn't find anything */
    return 0;
}


/*
 *   Get the line link record for a given source line 
 */
CHtmlDbg_line_link *CHtmlDebugHelper::find_line_link(
    int line_source_id, unsigned long seekpos)
{
    CHtmlDbg_line_link *link;

    /* find the line in our list */
    for (link = first_line_ ; link != 0 ; link = link->next_)
    {
        /* see if this is the one we're looking for */
        if (link->equals(line_source_id, seekpos))
        {
            /* this is it - return it */
            return link;
        }
    }

    /* didn't find it */
    return 0;
}


/*
 *   Get information on the source line containing the given breakpoint 
 */
int CHtmlDebugHelper::get_bp_line_info(dbgcxdef *dbgctx, int bpnum,
                                       int *source_id,
                                       unsigned long *seekpos)
{
    char desc[HTMLDBG_MAXBPNAME];
    
    /* get information on the specified breakpoint */
    if (dbgbpgetinfo(dbgctx, bpnum, desc, sizeof(desc), 0, 0))
        return FALSE;

    /* parse the descriptor */
    switch(desc[0])
    {
    case '#':
        /* it's a source line - parse the offset and filename */
        {
            char *p;
            lindef *lin;

            /* find the filename portion */
            for (p = desc + 1 ; *p != '\0' && *p != ' ' ; ++p) ;

            /* read the offset */
            *seekpos = get_atol(desc + 1);

            /* find the line source */
            for (lin = dbgctx->dbgcxlin ; lin != 0 ; lin = lin->linnxt)
            {
                /* if this is the one, return it */
                if (get_strcmp(p + 1, ((linfdef *)lin)->linfnam) == 0)
                {
                    *source_id = lin->linid;
                    break;
                }
            }

            /* if we didn't find the line source, return failure */
            if (lin == 0)
                return FALSE;
        }

        /* tell the caller we found the line */
        return TRUE;

    case 'g':
        /* 
         *   global breakpoint - there's no line information associated
         *   with these 
         */
        return FALSE;

    default:
        /* no source line associated with the breakpoint */
        return FALSE;
    }
}

/*
 *   Set or clear a source line status bit
 */
void CHtmlDebugHelper::change_line_status(int line_source_id,
                                          unsigned long seekpos,
                                          unsigned int statflag, int set)
{
    CHtmlDbg_line_link *link;

    /* find the link */
    link = find_line_link(line_source_id, seekpos);

    /* if we found it, adjust the current flags as requested */
    if (link != 0)
    {
        /* set or clear the new flags as needed */
        if (set)
            link->stat_ |= statflag;
        else
            link->stat_ &= ~statflag;
        
        /* update the line on-screen */
        update_line_on_screen(link);
        
        /* 
         *   if the flags are now all cleared in this line, drop the line
         *   from the special lines list, since there's nothing special to
         *   track about it any more 
         */
        if (link->stat_ == 0)
        {
            /* unlink it from the list */
            if (link->prev_ == 0)
                first_line_ = link->next_;
            else
                link->prev_->next_ = link->next_;
            if (link->next_ == 0)
                last_line_ = link->prev_;
            else
                link->next_->prev_ = link->prev_;
            
            /* delete it */
            delete link;
        }
    }
    else
    {
        /*
         *   We didn't find the line we were looking for, so the line
         *   doesn't currently have any special properties.  If we're
         *   clearing bits, there's nothing to do, since the line already
         *   has all its status bits cleared.  
         */
        if (!set)
            return;
        
        /* create a new special-line link and add it to the list */
        link = new CHtmlDbg_line_link(line_source_id, seekpos, statflag);
        link->prev_ = last_line_;
        link->next_ = 0;
        if (last_line_ != 0)
            last_line_->next_ = link;
        else
            first_line_ = link;
        last_line_ = link;
        
        /* update the line on-screen */
        update_line_on_screen(link);
    }
}

/*
 *   Update a line on-screen 
 */
void CHtmlDebugHelper::update_line_on_screen(CHtmlDbg_line_link *link)
{
    CHtmlDbg_win_link *win;
    
    /*
     *   Find the line's file in our file list.  If the line isn't in our
     *   file list anywhere, it's not displayed anywhere on screen, so
     *   there's nothing to update just now. 
     */
    for (win = first_srcwin_ ; win != 0 ; win = win->next_)
    {
        /* 
         *   See if this window's line source ID matches the line's source
         *   ID.  Note that we'll keep going even if we find a match,
         *   since we could conceivably have the same file open in
         *   multiple windows.  
         */
        if (win->source_id_ == link->source_id_)
            update_line_on_screen(win, link);
    }
}

/*
 *   update a particular line in a particular window on-screen 
 */
void CHtmlDebugHelper::update_line_on_screen(CHtmlDbg_win_link *win,
                                             CHtmlDbg_line_link *link)
{
    CHtmlDispTextDebugsrc *disp;

    /* find the display item for this line */
    disp = (CHtmlDispTextDebugsrc *)win->formatter_->
           find_by_debugsrcpos(link->seekpos_);

    /* if we found it, update its status */
    if (disp != 0)
        disp->set_dbg_status(win->win_, link->stat_);
}


/*
 *   Context structure for the stack iterator callback 
 */
struct HtmlDbg_stkdisp_t
{
    HtmlDbg_stkdisp_t(CHtmlDebugHelper *helper)
    {
        /* remember the helper object */
        helper_ = helper;

        /* start out at level 0 */
        level_ = 0;
    }
    
    /* the helper object */
    CHtmlDebugHelper *helper_;

    /* stack level - we increment this for each line in the stack trace */
    int level_;

    /* buffer for batching up each line of text */
    CStringBuf buf_;
};

/*
 *   Update the stack window 
 */
void CHtmlDebugHelper::update_stack_window(dbgcxdef *ctx,
                                           int reset_level)
{
    HtmlDbg_stkdisp_t cbctx(this);
    static textchar_t emptymsg[] = "[stack trace unavailable]";
    CHtmlDispTextDebugsrc *disp;
    int level;
    
    /* if we don't have a stack window, there's nothing to update */
    if (stack_win_ == 0)
        return;

    /* determine what level we'll be at when we're done */
    if (reset_level)
    {
        /* we're resetting the level to zero */
        level = 0;
    }
    else
    {
        /* remember the current level */
        level = (ctx_stackwin_disp_ == 0
                 ? 0 : ctx_stackwin_disp_->get_debugsrcpos());
    }
    
    /* 
     *   make sure we're not holding on to a context display item, since
     *   we're about to delete all of the display items in the stack
     *   window 
     */
    ctx_stackwin_disp_ = 0;
    
    /* delete everything currently in the stack window */
    stack_win_->formatter_->start_at_top(TRUE);
    stack_win_->formatter_->set_font(stack_win_->font_);
    stack_win_->parser_->clear_page();

    /* build the stack */
    dbgstktr(ctx, &update_stack_disp_cb, &cbctx, 0, FALSE, FALSE);

    /*
     *   If there's nothing on the stack at all, it means the game is
     *   running.  Put an indication of this in the stack listing so that
     *   it's not completely empty. 
     */
    if (cbctx.level_ == 0)
        stack_win_->parser_->append_tag(new CHtmlTagTextDebugsrc(
            0, tab_size_, stack_win_->parser_->get_text_array(),
            emptymsg, get_strlen(emptymsg)));

    /* add a blank line at the end */
    stack_win_->parser_->append_tag(new CHtmlTagTextDebugsrc(
        0, tab_size_, stack_win_->parser_->get_text_array(), "", 0));

    /* format the stack into the stack window */
    if (!stack_win_->win_->do_formatting(TRUE, TRUE, TRUE))
    {
        CHtmlRect area(0, 0, HTMLSYSWIN_MAX_RIGHT, HTMLSYSWIN_MAX_BOTTOM);
        stack_win_->win_->inval_doc_coords(&area);
    }

    /* update the scrollbars */
    stack_win_->win_->fmt_adjust_vscroll();

    /* 
     *   Turn on the current-line status flag for the first item - the
     *   first item in the list is always the current line, so display it
     *   thus.  Don't do this if there's no stack traceback available.  
     */
    disp = (CHtmlDispTextDebugsrc *)
           stack_win_->formatter_->find_by_debugsrcpos(0);
    if (disp != 0)
        disp->set_dbg_status(stack_win_->win_,
                             cbctx.level_ == 0 ? 0 :
                             HTMLTDB_STAT_CURLINE);

    /* set the context line, if we're at a nested context */
    if (level != 0)
    {
        /* get the correct level */
        ctx_stackwin_disp_ = (CHtmlDispTextDebugsrc *)
            stack_win_->formatter_->find_by_debugsrcpos(level);

        /* set its status */
        if (ctx_stackwin_disp_ != 0)
            ctx_stackwin_disp_->set_dbg_status(stack_win_->win_,
                                               HTMLTDB_STAT_CTXLINE);
    }
}

/*
 *   Display callback for for update_stack_window 
 */
void CHtmlDebugHelper::update_stack_disp_cb(void *ctx0, char *str, int strl)
{
    HtmlDbg_stkdisp_t *ctx = (HtmlDbg_stkdisp_t *)ctx0;
    
    /* invoke the non-static callback handler */
    ctx->helper_->update_stack_disp(ctx, str, strl);
}

void CHtmlDebugHelper::update_stack_disp(HtmlDbg_stkdisp_t *ctx,
                                         char *str, int strl)
{
    CHtmlTextBuffer txtbuf;
    char *p;

    /* look for a newline */
    for (p = str ; *p != '\0' ; ++p)
    {
        /* if this is a newline, add the line to the window */
        if (*p == '\n')
        {
            CHtmlTag *tag;

            /* if we have anything prior to the newline, add to the window */
            if (p != str)
            {
                /* append up to the newline to our line buffer */
                ctx->buf_.append(str, p - str);
            }

            /* 
             *   Add the line in the buffer to our list.  Note that we use
             *   the line number (i.e., the stack nesting level) as the
             *   position. 
             */
            tag = new CHtmlTagTextDebugsrc
                  (ctx->level_, tab_size_,
                   stack_win_->parser_->get_text_array(),
                   ctx->buf_.get(), get_strlen(ctx->buf_.get()));
            stack_win_->parser_->append_tag(tag);

            /* count the line */
            ctx->level_++;

            /* the buffer has been used, so clear it */
            ctx->buf_.set("");
            
            /* drop everything prior to the newline */
            str = p + 1;
        }
    }
    
    /* add the remainder of the line to the buffer */
    if (*str != '\0')
        ctx->buf_.append(str);
}

/*
 *   Activate a stack window item 
 */
CHtmlSysWin *CHtmlDebugHelper::
   activate_stack_win_item(CHtmlDisp *disp, dbgcxdef *ctx,
                           CHtmlDebugSysIfc_win *syswinifc)
{
    CHtmlDispTextDebugsrc *dbgdisp = (CHtmlDispTextDebugsrc *)disp;
    unsigned long level;
    
    /* forget any existing context line we have */
    forget_ctx_line();

    /* 
     *   get the level from the item - we store this as the file position
     *   of the item 
     */
    level = dbgdisp->get_debugsrcpos();

    /* activate the level */
    return activate_stack_win_line(level, ctx, syswinifc);
}

/*
 *   Activate a stack window line 
 */
CHtmlSysWin *CHtmlDebugHelper::
   activate_stack_win_line(int level, dbgcxdef *ctx,
                           CHtmlDebugSysIfc_win *syswinifc)
{
    /* 
     *   if it's level 0, select the current line; otherwise, activate the
     *   enclosing stack context level 
     */
    if (level == 0)
    {
        /* no new context line - activate the current line */
        return load_current_source(ctx, syswinifc);
    }
    else
    {
        CHtmlSysWin *win;
        CHtmlDispTextDebugsrc *dbgdisp;

        /* load the source at this level */
        win = load_source_at_level(ctx, syswinifc, level);

        /* update the stack window context display if possible */
        if (stack_win_ != 0)
        {
            /* find the display item for this line */
            dbgdisp = (CHtmlDispTextDebugsrc *)
                      stack_win_->formatter_->get_disp_by_linenum(level);
            
            /* set the stack window to show that the new context line */
            dbgdisp->set_dbg_status(stack_win_->win_, HTMLTDB_STAT_CTXLINE);
            ctx_stackwin_disp_ = dbgdisp;
        }
            
        /* return the window */
        return win;
    }
}

/*
 *   Update the history window 
 */
void CHtmlDebugHelper::update_hist_window(dbgcxdef *ctx)
{
    const textchar_t *startp;
    const textchar_t *endp;
    const textchar_t *p;

    /* if there's no history window, there's nothing to update */
    if (hist_win_ == 0)
        return;

    /* delete everything currently in the window */
    hist_win_->formatter_->start_at_top(TRUE);
    hist_win_->formatter_->set_font(hist_win_->font_);
    hist_win_->parser_->clear_page();
    
    /* run through the history buffer */
    startp = get_call_trace_buf(ctx);
    endp = startp + get_call_trace_len(ctx);
    for (p = startp ; p < endp ; p += get_strlen(p) + 1)
    {
        CHtmlTag *tag;
        
        /* store this line */
        tag = new CHtmlTagTextDebugsrc
              (0, tab_size_, hist_win_->parser_->get_text_array(),
               p, get_strlen(p));
        hist_win_->parser_->append_tag(tag);
    }

    /* add a blank line at the end */
    hist_win_->parser_->append_tag(new CHtmlTagTextDebugsrc(
        0, tab_size_, hist_win_->parser_->get_text_array(), "", 0));

    /* format it into the window */
    if (!hist_win_->win_->do_formatting(TRUE, TRUE, TRUE))
    {
        CHtmlRect area(0, 0, HTMLSYSWIN_MAX_RIGHT, HTMLSYSWIN_MAX_BOTTOM);
        hist_win_->win_->inval_doc_coords(&area);
    }

    /* update the scrollbars */
    hist_win_->win_->fmt_adjust_vscroll();
}

/*
 *   Toggle hidden output 
 */
void CHtmlDebugHelper::toggle_dbg_hid()
{
    dbghid = !dbghid;
}

/*
 *   Get hidden output display status 
 */
int CHtmlDebugHelper::get_dbg_hid()
{
    return dbghid;
}

/*
 *   Enumerate line sources 
 */
void CHtmlDebugHelper::
   enum_source_files(dbgcxdef *dbgctx,
                     void (*func)(void *cbctx, const char *fname,
                                  int line_source_id), void *cbctx)
{
    lindef *lin;
    
    /* run through the line sources, and invoke the callback for each one */
    for (lin = dbgctx->dbgcxlin ; lin != 0 ; lin = lin->linnxt)
        (*func)(cbctx, (textchar_t *)((linfdef *)lin)->linfnam, lin->linid);
}

/* ------------------------------------------------------------------------ */
/*
 *   CHtmlDbg_win_link implementation 
 */

CHtmlDbg_win_link::CHtmlDbg_win_link(CHtmlSysWin *win, CHtmlParser *parser,
                                     CHtmlFormatter *formatter, linfdef *linf,
                                     const textchar_t *win_title,
                                     CHtmlSysFont *font,
                                     HtmlDbg_wintype_t win_type)
{
    /* remember window and parser */
    win_ = win;
    parser_ = parser;
    formatter_ = formatter;

    /* not in a list yet, so clear link pointers */
    next_ = prev_ = 0;

    /* remember the name of the file */
    if (linf != 0)
        fname_.set((textchar_t *)linf->linfnam);
    else
        fname_.set(win_title);

    /* remember the line source ID */
    source_id_ = (linf == 0 ? -1 : linf->linflin.linid);

    /* remember the default font */
    font_ = font;

    /* remember the type of the window */
    win_type_ = win_type;
}

/*
 *   Find the line source for a given line source ID 
 */
lindef *CHtmlDbg_win_link::find_lindef(struct dbgcxdef *ctx, int id)
{
    lindef *lin;

    /* scan the list of line sources */
    for (lin = ctx->dbgcxlin ; lin != 0 ; lin = lin->linnxt)
    {
        /* if this is the one, return it */
        if (lin->linid == id)
            return lin;
    }

    /* not found */
    return 0;
}

#if 0 // old routines - not currently used - keep for now in case
      // we need them again
/*
 *   determine if my filename matches that of a linf 
 */
int CHtmlDbg_win_link::fname_equals(linfdef *linf)
{
    /* return true if the names match */
    return (get_strcmp((textchar_t *)linf->linfnam, fname_.get()) == 0);
}

/*
 *   Determine if I match a linf.  If the line source ID is not -1, it
 *   indicates that the linf came from the debug records, so we simply
 *   match line source ID's.  If the line source ID is -1, it means that
 *   the linf came from a random file opened by the user, so we compare
 *   filenames.  
 */
int CHtmlDbg_win_link::linf_equals(linfdef *linf)
{
    /* if the line source ID is valid, compare source ID's */
    if (linf->linflin.linid != -1)
        return (linf->linflin.linid == source_id_);

    /* the source ID is invalid, so simply compare filenames */
    return fname_equals(linf);
}
#endif /* 0 $$$ */

/* ------------------------------------------------------------------------ */
/*
 *   Special-line record 
 */

CHtmlDbg_line_link::CHtmlDbg_line_link(int source_id, unsigned long seekpos,
                                       unsigned int init_stat)
{
    /* remember the information */
    source_id_ = source_id;
    seekpos_ = seekpos;
    stat_ = init_stat;

    /* not in a list yet */
    next_ = prev_ = 0;
}

