#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/html/win32/htmlpref.cpp,v 1.1.1.1 1999/05/09 21:32:31 MJRoberts Exp $";
#endif

/* Copyright (c) 1997 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  htmlpref.cpp - preferences dialog
Function
  
Notes
  
Modified
  10/26/97 MJRoberts  - Creation
*/

#include <windows.h>
#include <commctrl.h>

#ifndef TADSHTML_H
#include "tadshtml.h"
#endif
#ifndef TADSAPP_H
#include "tadsapp.h"
#endif
#ifndef HTMLRES_H
#include "htmlres.h"
#endif
#ifndef HTMLPREF_H
#include "htmlpref.h"
#endif
#ifndef TADSDLG_H
#include "tadsdlg.h"
#endif
#ifndef HTMLW32_H
#include "htmlw32.h"
#endif
#ifndef TADSREG_H
#include "tadsreg.h"
#endif
#ifndef TADSCBTN_H
#include "tadscbtn.h"
#endif


/* ------------------------------------------------------------------------ */
/*
 *   Generic font-popup dialog - this class has some simple routines for
 *   handling property page dialogs with popups. 
 */
class CHtmlDialogFontPp: public CTadsDialogPropPage
{
public:
    CHtmlDialogFontPp(CHtmlPreferences *prefs) { prefs_ = prefs; }
    
protected:
    /* 
     *   Read the value of one of the font popups; returns true if this
     *   changes the font setting.  If 'store' is false, just checks the
     *   value without storing it.  
     */
    int read_font_popup(int id, HTML_pref_id_t prefid, int store);

    /* the main preferences object */
    CHtmlPreferences *prefs_;
};


/*
 *   read the contents of one of the font popups into a CStringBuf 
 */
int CHtmlDialogFontPp::read_font_popup(int id, HTML_pref_id_t prefid,
                                       int store)
{
    HWND ctl;
    char buf[128];
    size_t len;
    int idx;
    int changed;
    const textchar_t *oldval;

    /* get the control handle */
    ctl = GetDlgItem(handle_, id);

    /* get the selection number */
    idx = SendMessage(ctl, CB_GETCURSEL, 0, 0);

    /* if it's not a valid selection, don't make any changes */
    if (idx == CB_ERR)
        return FALSE;

    /* get the selected string */
    len = SendMessage(ctl, CB_GETLBTEXT, (WPARAM)idx, (LPARAM)buf);

    /* note whether this is a change */
    oldval = prefs_->get_val_str(prefid);
    changed = (oldval == 0 || len != get_strlen(oldval)
               || memcmp(buf, oldval, len) != 0);

    /* store it in the specified preference property */
    if (store)
        prefs_->set_val_str(prefid, buf, len);

    /* return change indication */
    return changed;
}

/* ------------------------------------------------------------------------ */
/*
 *   Font dialog 
 */


class CHtmlDialogFonts: public CHtmlDialogFontPp
{
public:
    CHtmlDialogFonts(CHtmlPreferences *prefs)
        : CHtmlDialogFontPp(prefs) { }
    void init();
    int do_notify(NMHDR *nm);
    int do_command(WPARAM cmd, HWND ctl);

private:
    /* callbacks to select specific parameterized font types */
    static int font_select_serif(void *, ENUMLOGFONT *, NEWTEXTMETRIC *);
    static int font_select_sans(void *, ENUMLOGFONT *, NEWTEXTMETRIC *);
    static int font_select_script(void *, ENUMLOGFONT *, NEWTEXTMETRIC *);
    static int font_select_typewriter(void *, ENUMLOGFONT *, NEWTEXTMETRIC *);

    /* input text color */
    HTML_color_t input_text_color_;

    /* special popup entry for "Main text font" */
    static const char *main_font_entry;
};

/* main font entry text (for special input font name selection) */
const char *CHtmlDialogFonts::main_font_entry = "(Main Game Font)";

/*
 *   initialize 
 */
void CHtmlDialogFonts::init()
{
    COLORREF *cust_colors = prefs_->get_cust_colors();
    const char *cur_inp_font;
    HWND pop;

    /* set up the font combos with lists of available fonts */
    init_font_popup(IDC_MAINFONTPOPUP, TRUE, TRUE,
                    prefs_->get_prop_font(), TRUE);
    init_font_popup(IDC_MONOFONTPOPUP, FALSE, TRUE,
                    prefs_->get_mono_font(), TRUE);
    init_font_popup(IDC_FONTSERIF, prefs_->get_font_serif(),
                    &font_select_serif, 0, TRUE);
    init_font_popup(IDC_FONTSANS, prefs_->get_font_sans(),
                    &font_select_sans, 0, TRUE);
    init_font_popup(IDC_FONTSCRIPT, prefs_->get_font_script(),
                    &font_select_script, 0, TRUE);
    init_font_popup(IDC_FONTTYPEWRITER, prefs_->get_font_typewriter(),
                    &font_select_typewriter, 0, TRUE);

    /* get the current input font from the preferences */
    cur_inp_font = prefs_->get_inpfont_name();

    /* if the name is empty, use the "Main Text Font" entry by default */
    if (cur_inp_font == 0 || cur_inp_font[0] == '\0')
        cur_inp_font = main_font_entry;

    /* get the font selector control */
    pop = GetDlgItem(handle_, IDC_POP_INPUTFONT);

    /* clear out any existing list contents */
    SendMessage(pop, CB_RESETCONTENT, 0, 0);

    /* add an element to the font popup for "Main text font" */
    SendMessage(pop, CB_ADDSTRING, (WPARAM)0, (LPARAM)main_font_entry);

    /* set up the font combo with a list of the available fonts */
    init_font_popup(IDC_POP_INPUTFONT, TRUE, TRUE, cur_inp_font, FALSE);

    /* add the color button */
    input_text_color_ = prefs_->get_inpfont_color();
    controls_.add(new CColorBtnPropPage(handle_, IDC_BTN_INPUTCOLOR,
                                        &input_text_color_,
                                        cust_colors, this));

    /* initialize the input font bold/italic style settings */
    CheckDlgButton(handle_, IDC_CK_INPUTBOLD,
                   (prefs_->get_inpfont_bold()
                    ? BST_CHECKED : BST_UNCHECKED));
    CheckDlgButton(handle_, IDC_CK_INPUTITALIC,
                   (prefs_->get_inpfont_italic()
                    ? BST_CHECKED : BST_UNCHECKED));
}

/*
 *   select serifed fonts 
 */
int CHtmlDialogFonts::font_select_serif(void *, ENUMLOGFONT *elf,
                                       NEWTEXTMETRIC *tm)
{
    /* return ROMAN-style fonts */
    return (tm->tmCharSet != SYMBOL_CHARSET
            && (elf->elfLogFont.lfPitchAndFamily & 0xf0) == FF_ROMAN);
}

/*
 *   select sans-serif fonts 
 */
int CHtmlDialogFonts::font_select_sans(void *, ENUMLOGFONT *elf,
                                       NEWTEXTMETRIC *tm)
{
    /* return SWISS-style fonts */
    return (tm->tmCharSet != SYMBOL_CHARSET
            && (elf->elfLogFont.lfPitchAndFamily & 0xf0) == FF_SWISS);
}

/*
 *   select script fonts 
 */
int CHtmlDialogFonts::font_select_script(void *, ENUMLOGFONT *elf,
                                         NEWTEXTMETRIC *tm)
{
    /* 
     *   return SCRIPT and ROMAN styles - a lot of script-type fonts seem
     *   to claim they're ROMAN fonts 
     */
    return (tm->tmCharSet != SYMBOL_CHARSET
            && ((elf->elfLogFont.lfPitchAndFamily & 0xf0) == FF_SCRIPT
                || (elf->elfLogFont.lfPitchAndFamily & 0xf0) == FF_ROMAN));
}

/*
 *   select typewriter fonts 
 */
int CHtmlDialogFonts::font_select_typewriter(void *, ENUMLOGFONT *elf,
                                             NEWTEXTMETRIC *tm)
{
    /* return MODERN-style fonts */
    return (tm->tmCharSet != SYMBOL_CHARSET
            && (elf->elfLogFont.lfPitchAndFamily & 0xf0) == FF_MODERN);
}

/*
 *   receive a notification message 
 */
int CHtmlDialogFonts::do_notify(NMHDR *nm)
{
    int ch;
    int new_bold, new_ital;

    switch(nm->code)
    {
    case PSN_APPLY:
        /* 
         *   read the popups - note that we explicitly read all of them
         *   separately rather than with if (a || b || ...), because we
         *   want to read them all - using || would short-circuit any
         *   after the first change, which we obviously don't want to do 
         */
        ch = read_font_popup(IDC_MAINFONTPOPUP, HTML_PREF_PROP_FONT, TRUE);
        ch |= read_font_popup(IDC_MONOFONTPOPUP, HTML_PREF_MONO_FONT, TRUE);
        ch |= read_font_popup(IDC_FONTSERIF, HTML_PREF_FONT_SERIF, TRUE);
        ch |= read_font_popup(IDC_FONTSANS, HTML_PREF_FONT_SANS, TRUE);
        ch |= read_font_popup(IDC_FONTSCRIPT, HTML_PREF_FONT_SCRIPT, TRUE);
        ch |= read_font_popup(IDC_FONTTYPEWRITER,
                              HTML_PREF_FONT_TYPEWRITER, TRUE);

        /* note the input font name, and check for a change */
        ch |= read_font_popup(IDC_POP_INPUTFONT,
                              HTML_PREF_INPFONT_NAME, TRUE);

        /* 
         *   if the special "Main Text Font" entry is selected, use an
         *   empty font name 
         */
        if (!strcmp(prefs_->get_inpfont_name(), main_font_entry))
            prefs_->set_inpfont_name("");

        /* check for a change in the color setting */
        ch |= (prefs_->get_inpfont_color() != input_text_color_);

        /* note the new style settings */
        new_bold = (IsDlgButtonChecked(handle_, IDC_CK_INPUTBOLD)
                    == BST_CHECKED);
        new_ital = (IsDlgButtonChecked(handle_, IDC_CK_INPUTITALIC)
                    == BST_CHECKED);

        /* check for style changes */
        ch |= ((new_bold != prefs_->get_inpfont_bold())
               || (new_ital != prefs_->get_inpfont_italic()));

        /* store the new input text color and style settings */
        prefs_->set_inpfont_bold(new_bold);
        prefs_->set_inpfont_italic(new_ital);
        prefs_->set_inpfont_color(input_text_color_);

        /* if anything changed, reformat the main window */
        if (ch)
        {
            /* font changes require updating the window */
            prefs_->schedule_reformat(FALSE);
        }

        /* handled */
        return TRUE;
    }

    return FALSE;
}

/*
 *   process a command 
 */
int CHtmlDialogFonts::do_command(WPARAM cmd, HWND ctl)
{
    /* see if it's a notification that one of the popups changed selection */
    if (HIWORD(cmd) == CBN_SELCHANGE)
    {
        /* see if anything changed */
        if (read_font_popup(IDC_MAINFONTPOPUP, HTML_PREF_PROP_FONT, FALSE)
            || read_font_popup(IDC_MONOFONTPOPUP, HTML_PREF_MONO_FONT, FALSE)
            || read_font_popup(IDC_FONTSERIF, HTML_PREF_FONT_SERIF, FALSE)
            || read_font_popup(IDC_FONTSANS, HTML_PREF_FONT_SANS, FALSE)
            || read_font_popup(IDC_FONTSCRIPT, HTML_PREF_FONT_SCRIPT, FALSE)
            || read_font_popup(IDC_FONTTYPEWRITER,
                               HTML_PREF_FONT_TYPEWRITER, FALSE)
            || read_font_popup(IDC_POP_INPUTFONT,
                               HTML_PREF_INPFONT_NAME, FALSE))
        {
            /* tell the page that something's changed */
            set_changed(TRUE);
        }

        /* handled */
        return TRUE;
    }

    /* check for other commands */
    switch(cmd)
    {
    case IDC_CK_INPUTBOLD:
    case IDC_CK_INPUTITALIC:
        /* note the state change */
        set_changed(TRUE);
        return TRUE;

    default:
        /* inherit the default behavior */
        return CHtmlDialogFontPp::do_command(cmd, ctl);
    }
}


/* ------------------------------------------------------------------------ */
/*
 *   Color dialog
 */

class CHtmlDialogColor: public CTadsDialogPropPage
{
public:
    CHtmlDialogColor(CHtmlPreferences *prefs) { prefs_ = prefs; }
    void init();
    int do_command(WPARAM cmd, HWND ctl);
    int do_notify(NMHDR *nm);

private:
    CHtmlPreferences *prefs_;

    /* current color settings */
    HTML_color_t bg_color_;
    HTML_color_t text_color_;
    HTML_color_t link_color_;
    HTML_color_t vlink_color_;
    HTML_color_t hlink_color_;
    HTML_color_t alink_color_;
    HTML_color_t stat_text_color_;
    HTML_color_t stat_bg_color_;
};


/* 
 *   initialize 
 */
void CHtmlDialogColor::init()
{
    int use_win_colors;
    int override_colors;
    int hover_hilite;
    COLORREF *cust_colors = prefs_->get_cust_colors();
    
    /* get initial colors from preferences */
    bg_color_ = prefs_->get_bg_color();
    text_color_ = prefs_->get_text_color();
    link_color_ = prefs_->get_link_color();
    vlink_color_ = prefs_->get_vlink_color();
    hlink_color_ = prefs_->get_hlink_color();
    alink_color_ = prefs_->get_alink_color();
    stat_text_color_ = prefs_->get_color_status_text();
    stat_bg_color_ = prefs_->get_color_status_bg();

    /* set up the color buttons */
    controls_.add(new CColorBtnPropPage(handle_, IDC_BKCOLOR,
                                        &bg_color_, cust_colors, this));
    controls_.add(new CColorBtnPropPage(handle_, IDC_TXTCOLOR,
                                        &text_color_, cust_colors, this));
    controls_.add(new CColorBtnPropPage(handle_, IDC_LINKCOLOR,
                                        &link_color_, cust_colors, this));
    controls_.add(new CColorBtnPropPage(handle_, IDC_HLINKCOLOR,
                                        &hlink_color_, cust_colors, this));
    controls_.add(new CColorBtnPropPage(handle_, IDC_ALINKCOLOR,
                                        &alink_color_, cust_colors, this));
    controls_.add(new CColorBtnPropPage(handle_, IDC_SBKCOLOR,
                                        &stat_bg_color_, cust_colors, this));
    controls_.add(new CColorBtnPropPage(handle_, IDC_STXTCOLOR,
                                        &stat_text_color_,
                                        cust_colors, this));

    /* initialize enabling state of text/background color buttons */
    use_win_colors = prefs_->get_use_win_colors();
    override_colors = prefs_->get_override_colors();
    CheckDlgButton(handle_, IDC_CK_USEWINCLR,
                   (use_win_colors ? BST_CHECKED : BST_UNCHECKED));
    CheckDlgButton(handle_, IDC_CK_OVERRIDECLR,
                   (override_colors ? BST_CHECKED : BST_UNCHECKED));

    /* enable or disable the text and background buttons accordingly */
    EnableWindow(GetDlgItem(handle_, IDC_TXTCOLOR), !use_win_colors);
    EnableWindow(GetDlgItem(handle_, IDC_BKCOLOR), !use_win_colors);

    /* initialize underline checkbox */
    CheckDlgButton(handle_, IDC_CK_UNDERLINE,
                   (prefs_->get_underline_links()
                    ? BST_CHECKED : BST_UNCHECKED));

    /* initialize hover hilite checkbox */
    hover_hilite = prefs_->get_hover_hilite();
    CheckDlgButton(handle_, IDC_CK_HOVER,
                   (hover_hilite ? BST_CHECKED : BST_UNCHECKED));

    /* enable or disable the hover color accordingly */
    EnableWindow(GetDlgItem(handle_, IDC_HLINKCOLOR), hover_hilite);
}

/*
 *   process notifications 
 */
int CHtmlDialogColor::do_notify(NMHDR *nm)
{
    int new_use_win_clr;
    int new_override_clr;
    int new_underline;
    int new_hover_hilite;
    
    switch(nm->code)
    {
    case PSN_APPLY:
        /* note new checkbox settings */
        new_use_win_clr = (IsDlgButtonChecked(handle_, IDC_CK_USEWINCLR)
                           == BST_CHECKED);
        new_override_clr = (IsDlgButtonChecked(handle_, IDC_CK_OVERRIDECLR)
                            == BST_CHECKED);
        new_underline = (IsDlgButtonChecked(handle_, IDC_CK_UNDERLINE)
                         == BST_CHECKED);
        new_hover_hilite = (IsDlgButtonChecked(handle_, IDC_CK_HOVER)
                            == BST_CHECKED);

        /* if any of the settings changed, we need to reformat the window */
        if (prefs_->get_use_win_colors() != new_use_win_clr
            || prefs_->get_override_colors() != new_override_clr
            || prefs_->get_bg_color() != bg_color_
            || prefs_->get_text_color() != text_color_
            || prefs_->get_link_color() != link_color_
            || prefs_->get_vlink_color() != vlink_color_
            || prefs_->get_hlink_color() != hlink_color_
            || prefs_->get_alink_color() != alink_color_
            || prefs_->get_underline_links() != new_underline
            || prefs_->get_hover_hilite() != new_hover_hilite
            || prefs_->get_color_status_text() != stat_text_color_
            || prefs_->get_color_status_bg() != stat_bg_color_)
            prefs_->schedule_reformat(FALSE);
        
        /* store the current color settings in the preferences */
        prefs_->set_use_win_colors(new_use_win_clr);
        prefs_->set_override_colors(new_override_clr);
        prefs_->set_bg_color(bg_color_);
        prefs_->set_text_color(text_color_);
        prefs_->set_link_color(link_color_);
        prefs_->set_vlink_color(vlink_color_);
        prefs_->set_hlink_color(hlink_color_);
        prefs_->set_alink_color(alink_color_);
        prefs_->set_underline_links(new_underline);
        prefs_->set_hover_hilite(new_hover_hilite);
        prefs_->set_color_status_text(stat_text_color_);
        prefs_->set_color_status_bg(stat_bg_color_);
        return TRUE;
    }
    return FALSE;
}

/*
 *   process a command 
 */
int CHtmlDialogColor::do_command(WPARAM id, HWND ctl)
{
    switch(id)
    {
    case IDC_CK_USEWINCLR:
        {
            int checked;
            
            /* get the state of the button */
            checked = (IsDlgButtonChecked(handle_, (int)id) == BST_CHECKED);
            
            /* 
             *   enable or disable the text and background buttons
             *   accordingly 
             */
            EnableWindow(GetDlgItem(handle_, IDC_TXTCOLOR), !checked);
            EnableWindow(GetDlgItem(handle_, IDC_BKCOLOR), !checked);

            /* note the state change */
            set_changed(TRUE);
        }
        return TRUE;

    case IDC_CK_HOVER:
        {
            int checked;

            /* get the state */
            checked = (IsDlgButtonChecked(handle_, (int)id) == BST_CHECKED);

            /* enable or disable the hover color accordingly */
            EnableWindow(GetDlgItem(handle_, IDC_HLINKCOLOR), checked);

            /* note the state change */
            set_changed(TRUE);
        }
        return TRUE;
        
    case IDC_CK_UNDERLINE:
    case IDC_CK_OVERRIDECLR:
        /* note the state change */
        set_changed(TRUE);
        return TRUE;
        
    default:
        /* pass to superclass for processing */
        return CTadsDialog::do_command(id, ctl);
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Keys dialog 
 */

class CHtmlDialogKeys: public CTadsDialogPropPage
{
public:
    CHtmlDialogKeys(CHtmlPreferences *prefs) { prefs_ = prefs; }
    void init();
    int do_notify(NMHDR *nm);
    int do_command(WPARAM cmd, HWND ctl);
    
private:
    /* get the button states */
    void get_button_state(int *emacs_ctrl_v, int *arrow_scroll,
                          int *emacs_alt_v);

    CHtmlPreferences *prefs_;
};

/*
 *   initialize 
 */
void CHtmlDialogKeys::init()
{
    /* initialize controls with current preference settings */
    CheckRadioButton(handle_, IDC_RB_CTLV_EMACS, IDC_RB_CTLV_PASTE,
                     prefs_->get_emacs_ctrl_v()
                     ? IDC_RB_CTLV_EMACS : IDC_RB_CTLV_PASTE);
    CheckRadioButton(handle_, IDC_RB_ARROW_SCROLL, IDC_RB_ARROW_HIST,
                     prefs_->get_arrow_keys_always_scroll()
                     ? IDC_RB_ARROW_SCROLL : IDC_RB_ARROW_HIST);
    CheckRadioButton(handle_, IDC_RB_ALTV_EMACS, IDC_RB_ALTV_WIN,
                     prefs_->get_emacs_alt_v()
                     ? IDC_RB_ALTV_EMACS : IDC_RB_ALTV_WIN);
}

/*
 *   process notifications 
 */
int CHtmlDialogKeys::do_notify(NMHDR *nm)
{
    switch(nm->code)
    {
    case PSN_APPLY:
        {
            int new_arrow, new_emacs_ctl, new_emacs_alt;

            /* get the current button state */
            get_button_state(&new_emacs_ctl, &new_arrow, &new_emacs_alt);

            /* save the current selection state in the preferences */
            prefs_->set_emacs_ctrl_v(new_emacs_ctl);
            prefs_->set_arrow_keys_always_scroll(new_arrow);
            prefs_->set_emacs_alt_v(new_emacs_alt);

            /* handled */
            return TRUE;
        }
    }
    return FALSE;
}

/*
 *   get the current state 
 */
void CHtmlDialogKeys::get_button_state(int *emacs_ctrl_v, int *arrow_scroll,
                                       int *emacs_alt_v)
{
    *emacs_ctrl_v = (IsDlgButtonChecked(handle_, IDC_RB_CTLV_EMACS)
                     == BST_CHECKED);
    *arrow_scroll = (IsDlgButtonChecked(handle_, IDC_RB_ARROW_SCROLL)
                     == BST_CHECKED);
    *emacs_alt_v = (IsDlgButtonChecked(handle_, IDC_RB_ALTV_EMACS)
                    == BST_CHECKED);
}

/*
 *   process commands 
 */
int CHtmlDialogKeys::do_command(WPARAM, HWND)
{
    int new_emacs_ctrl_v, new_arrow_scroll, new_emacs_alt_v;
    
    /* see if the state has changed */
    get_button_state(&new_emacs_ctrl_v, &new_arrow_scroll,
                     &new_emacs_alt_v);
    if ((new_emacs_ctrl_v != 0) != (prefs_->get_emacs_ctrl_v() != 0)
        || ((new_arrow_scroll != 0)
            != (prefs_->get_arrow_keys_always_scroll() != 0))
        || ((new_emacs_alt_v != 0) != (prefs_->get_emacs_alt_v() != 0)))
        set_changed(TRUE);

    /* handled */
    return TRUE;
}

/* ------------------------------------------------------------------------ */
/*
 *   MORE dialog 
 */

class CHtmlDialogMore: public CTadsDialogPropPage
{
public:
    CHtmlDialogMore(CHtmlPreferences *prefs) { prefs_ = prefs; }
    void init();
    int do_notify(NMHDR *nm);
    int do_command(WPARAM cmd, HWND ctl);

private:
    /* get the state of the Alt More Style button */
    int get_alt_more_button();

    CHtmlPreferences *prefs_;
};

/*
 *   initialize 
 */
void CHtmlDialogMore::init()
{
    /* initialize controls with current preference settings */
    CheckRadioButton(handle_, IDC_RB_MORE_NORMAL, IDC_RB_MORE_ALT,
                     prefs_->get_alt_more_style() ?
                     IDC_RB_MORE_ALT : IDC_RB_MORE_NORMAL);
}

/*
 *   process notifications 
 */
int CHtmlDialogMore::do_notify(NMHDR *nm)
{
    switch(nm->code)
    {
    case PSN_APPLY:
        /* 
         *   if the style has changed, and the window is in MORE mode, we
         *   need to reformat the window 
         */
        if (get_alt_more_button() != prefs_->get_alt_more_style())
            prefs_->schedule_reformat(TRUE);
        
        /* save the current selection state in the preferences */
        prefs_->set_alt_more_style(get_alt_more_button());
        return TRUE;
    }
    return FALSE;
}

/* 
 *   get the alt-more-mode button state 
 */
int CHtmlDialogMore::get_alt_more_button()
{
    return (IsDlgButtonChecked(handle_, IDC_RB_MORE_ALT) == BST_CHECKED);
}

/*
 *   process commands 
 */
int CHtmlDialogMore::do_command(WPARAM, HWND)
{
    /* see if anything has changed */
    if ((prefs_->get_alt_more_style() != 0) != (get_alt_more_button() != 0))
        set_changed(TRUE);

    /* handled */
    return TRUE;
}

/* ------------------------------------------------------------------------ */
/*
 *   File safety level dialog 
 */

class CHtmlDialogSafety: public CTadsDialogPropPage
{
public:
    CHtmlDialogSafety(CHtmlPreferences *prefs)
    {
        /* save the preferences object */
        prefs_ = prefs;

        /* remember whether we've been changed by the user */
        user_change_ = FALSE;
    }
    void init();
    int do_notify(NMHDR *nm);
    int do_command(WPARAM cmd, HWND ctl);

private:
    /* flag indicating that the user has explicitly changed our setting */
    int user_change_ : 1;

    /* 
     *   array of our button identifiers; the array index is the
     *   corresponding safety level setting 
     */
    static int buttons[];

    /* preferences object */
    CHtmlPreferences *prefs_;
};

int CHtmlDialogSafety::buttons[] =
{
    IDC_RB_SAFETY0,
    IDC_RB_SAFETY1,
    IDC_RB_SAFETY2,
    IDC_RB_SAFETY3,
    IDC_RB_SAFETY4
};

/*
 *   initialize 
 */
void CHtmlDialogSafety::init()
{
    /* initialize the appropriate radio button for the current setting */
    CheckRadioButton(handle_, IDC_RB_SAFETY0, IDC_RB_SAFETY4,
                     buttons[prefs_->get_file_safety_level()]);
}

/*
 *   handle a command 
 */
int CHtmlDialogSafety::do_command(WPARAM id, HWND ctl)
{
    switch(id)
    {
    case IDC_RB_SAFETY0:
    case IDC_RB_SAFETY1:
    case IDC_RB_SAFETY2:
    case IDC_RB_SAFETY3:
    case IDC_RB_SAFETY4:
        /* note the explicit user change in our own flag */
        user_change_ = TRUE;

        /* note it in the dialog as well */
        set_changed(TRUE);

        /* handled */
        return TRUE;

    default:
        /* inherit default handling */
        return CTadsDialogPropPage::do_command(id, ctl);
    }
}

/*
 *   process notifications 
 */
int CHtmlDialogSafety::do_notify(NMHDR *nm)
{
    switch(nm->code)
    {
    case PSN_APPLY:
        /* 
         *   if the user has explicitly changed one of our radio buttons,
         *   save the new state - don't save the state otherwise, since we
         *   may have a temporary setting from the command line that we
         *   don't want to store in the preferences 
         */
        if (user_change_)
        {
            int i;

            /* get the current button setting */
            for (i = 0 ; i < sizeof(buttons)/sizeof(buttons[0]) ; ++i)
            {
                /* if this button is checked, it's the new level */
                if (IsDlgButtonChecked(handle_, buttons[i]) == BST_CHECKED)
                {
                    /* save the new preference */
                    prefs_->set_file_safety_level(i);

                    /* no need to look further */
                    break;
                }
            }

            /* we've handled this change */
            user_change_ = FALSE;
        }

        /* handled */
        return TRUE;
    }

    /* not handled */
    return FALSE;
}


/* ------------------------------------------------------------------------ */
/*
 *   Memory dialog 
 */

class CHtmlDialogMem: public CTadsDialogPropPage
{
public:
    CHtmlDialogMem(CHtmlPreferences *prefs)
    {
        /* save the preferences object */
        prefs_ = prefs;
    }
    void init();
    int do_notify(NMHDR *nm);
    int do_command(WPARAM cmd, HWND ctl);

private:
    /* preferences object */
    CHtmlPreferences *prefs_;
};

/* 
 *   initialize 
 */
void CHtmlDialogMem::init()
{
    HWND pop;
    int i;
    static const char *settings[] =
    {
        "No Limit",
        "32 KBytes",
        "64 KBytes",
        "96 KBytes",
        "128 KBytes",
    };
    int settings_count;
    int idx;

    /* calculate the number of items in our list */
    settings_count  = sizeof(settings)/sizeof(settings[0]);
    
    /* add the possible settings to the popup */
    pop = GetDlgItem(handle_, IDC_POP_MEM);
    SendMessage(pop, CB_RESETCONTENT, 0, 0);
    for (i = 0 ; i < settings_count ; ++i)
        SendMessage(pop, CB_ADDSTRING, (WPARAM)0, (LPARAM)settings[i]);

    /* 
     *   Select the appropriate list item.  Get the current setting, and
     *   divide it by 32K to get the nearest appropriate index in our
     *   list.  If the current memory limit is zero, it means that we have
     *   no limit; our first element is conveniently "No Limit", hence a
     *   limit of zero is also an index of zero. 
     */
    idx = (int)(prefs_->get_mem_text_limit() / (unsigned long)32768);
    if (idx >= settings_count)
        idx = settings_count;

    /* select the list item */
    SendMessage(pop, CB_SETCURSEL, (WPARAM)idx, 0);
}

/*
 *   handle a command 
 */
int CHtmlDialogMem::do_command(WPARAM id, HWND ctl)
{
    if (HIWORD(id) == CBN_SELCHANGE)
    {
        /* note the change */
        set_changed(TRUE);

        /* handled */
        return TRUE;
    }

    /* not handled */
    return FALSE;
}

/*
 *   process notifications 
 */
int CHtmlDialogMem::do_notify(NMHDR *nm)
{
    int idx;
    
    switch(nm->code)
    {
    case PSN_APPLY:
        /* get the current selection in the text memory list */
        idx = SendMessage(GetDlgItem(handle_, IDC_POP_MEM),
                          CB_GETCURSEL, 0, 0);

        /* 
         *   If we got a valid index, set the new memory size in bytes.
         *   Each element of the list is a 32K increment, so multiply the
         *   index by 32768 to get the byte size.  Note that the first
         *   list item (at index 0) is "No Limit", hence our calculation
         *   will give us zero in this case; this is just what we want,
         *   since a value of zero bytes indicates unlimited memory. 
         */
        if (idx >= 0)
            prefs_->set_mem_text_limit((unsigned long)idx
                                       * (unsigned long)32768);

        /* handled */
        return TRUE;
    }

    /* not handled */
    return FALSE;
}




/* ------------------------------------------------------------------------ */
/*
 *   Preferences object
 */

/* registry key name */
const textchar_t CHtmlPreferences::reg_key_name[] =
    "Software\\TADS\\HTML TADS\\1.0\\Settings";

/* custom color value name */
const textchar_t CHtmlPreferences::custclr_val_name[] =
    "Custom Colors";


CHtmlPreferences::CHtmlPreferences()
{
    CHtmlRect rc;
    
    /* create the property list */
    proplist_ = new CHtmlPropList();

    /* initialize the property names (for serialization purposes) */
    proplist_->set_prop_name(HTML_PREF_PROP_FONT, "Proportional Font");
    proplist_->set_prop_name(HTML_PREF_MONO_FONT, "Monospaced Font");
    proplist_->set_prop_name(HTML_PREF_FONT_SERIF, "TADS-Serif Font");
    proplist_->set_prop_name(HTML_PREF_FONT_SANS, "TADS-Sans Font");
    proplist_->set_prop_name(HTML_PREF_FONT_SCRIPT, "TADS-Script Font");
    proplist_->set_prop_name(HTML_PREF_FONT_TYPEWRITER,
                             "TADS-Typewriter Font");
    proplist_->set_prop_name(HTML_PREF_FONT_SCALE, "Font Scale");
    proplist_->set_prop_name(HTML_PREF_USE_WIN_COLORS, "Use Windows Colors");
    proplist_->set_prop_name(HTML_PREF_OVERRIDE_COLORS, "Override Colors");
    proplist_->set_prop_name(HTML_PREF_TEXT_COLOR, "Text Color");
    proplist_->set_prop_name(HTML_PREF_BG_COLOR, "Background Color");
    proplist_->set_prop_name(HTML_PREF_LINK_COLOR, "Link Color");
    proplist_->set_prop_name(HTML_PREF_HLINK_COLOR, "Hovering Link Color");
    proplist_->set_prop_name(HTML_PREF_ALINK_COLOR, "Active Link Color");
    proplist_->set_prop_name(HTML_PREF_UNDERLINE_LINKS, "Underline Links");
    proplist_->set_prop_name(HTML_PREF_EMACS_CTRL_V, "Emacs Ctrl-V");
    proplist_->set_prop_name(HTML_PREF_EMACS_ALT_V, "Emacs Alt-V");
    proplist_->set_prop_name(HTML_PREF_ARROW_KEYS_ALWAYS_SCROLL,
                            "Arrows Keys Always Scroll");
    proplist_->set_prop_name(HTML_PREF_ALT_MORE_STYLE, "Alt More Style");
    proplist_->set_prop_name(HTML_PREF_WINPOS, "Window Position");
    proplist_->set_prop_name(HTML_PREF_DBGWINPOS, "Debug Window Position");
    proplist_->set_prop_name(HTML_PREF_DBGMAINPOS,
                             "Debug Main Window Position");
    proplist_->set_prop_name(HTML_PREF_MUSIC_ON, "Music");
    proplist_->set_prop_name(HTML_PREF_SOUNDS_ON, "Sound Effects");
    proplist_->set_prop_name(HTML_PREF_GRAPHICS_ON, "Graphics");
    proplist_->set_prop_name(HTML_PREF_LINKS_ON, "Show Links");
    proplist_->set_prop_name(HTML_PREF_LINKS_CTRL, "Show Links on Ctrl");
    proplist_->set_prop_name(HTML_PREF_DIRECTX_HIDEWARN,
                             "Hide DirectX Startup Warning");
    proplist_->set_prop_name(HTML_PREF_DIRECTX_ERROR_CODE,
                             "DirectX Initialization Result");
    proplist_->set_prop_name(HTML_PREF_FILE_SAFETY_LEVEL,
                             "File Safety Level");
    proplist_->set_prop_name(HTML_PREF_COLOR_STATUS_BG,
                             "Status line Background Color");
    proplist_->set_prop_name(HTML_PREF_COLOR_STATUS_TEXT,
                             "Status line Text Color");
    proplist_->set_prop_name(HTML_PREF_INPFONT_DEFAULT,
                             "Input Font Use Default");
    proplist_->set_prop_name(HTML_PREF_INPFONT_NAME,
                             "Input Font Name");
    proplist_->set_prop_name(HTML_PREF_INPFONT_COLOR,
                             "Input Font Color");
    proplist_->set_prop_name(HTML_PREF_INPFONT_BOLD,
                             "Input Font Bold");
    proplist_->set_prop_name(HTML_PREF_INPFONT_ITALIC,
                             "Input Font Italic");
    proplist_->set_prop_name(HTML_PREF_MEM_TEXT_LIMIT,
                             "Text Memory Limit");

    /* initialize default property values */
    set_prop_font("Times New Roman");
    set_mono_font("Courier New");
    set_font_serif("Times New Roman");
    set_font_sans("Arial");
    set_font_script("Times New Roman");
    set_font_typewriter("Courier New");
    set_font_scale(2);
    set_use_win_colors(TRUE);
    set_text_color(COLORREF_to_HTML_color(GetSysColor(COLOR_WINDOWTEXT)));
    set_bg_color(COLORREF_to_HTML_color(GetSysColor(COLOR_WINDOW)));
    set_link_color(HTML_make_color(0x00, 0x00, 0xff));
    set_vlink_color(HTML_make_color(0x00, 0x80, 0x80));
    set_alink_color(HTML_make_color(0xff, 0x00, 0xff));
    set_hlink_color(HTML_make_color(0xff, 0x00, 0x00));
    set_underline_links(TRUE);
    set_hover_hilite(FALSE);
    set_emacs_ctrl_v(FALSE);
    set_emacs_alt_v(FALSE);
    set_arrow_keys_always_scroll(FALSE);
    set_alt_more_style(TRUE);
    set_music_on(TRUE);
    set_sounds_on(TRUE);
    set_graphics_on(TRUE);
    set_links_on(TRUE);
    set_links_ctrl(FALSE);
    set_directx_hidewarn(FALSE);
    set_directx_error_code(HTMLW32_DIRECTX_OK);
    set_mem_text_limit(65536);

    /* use game font for input font by default */
    set_inpfont_name("");
    set_inpfont_color(HTML_make_color(0x00, 0x00, 0x00));
    set_inpfont_bold(FALSE);
    set_inpfont_italic(FALSE);

    /* 
     *   use black text on light gray ("silver," in html parlance)
     *   background for default status line color scheme 
     */
    set_color_status_text(0x000000);
    set_color_status_bg(0xc0c0c0);

    /* set up default initial window positions */
    rc.set(50, 50, 500, 600);
    set_win_pos(&rc);
    set_dbgwin_pos(&rc);

    rc.set(25, 25, 400, 50);
    set_dbgmain_pos(&rc);

    /* 
     *   By default, use the safety settings in the preferences (so don't
     *   use a temporary setting).  If it's not stored, default it to
     *   zero.  The run-time will set a temporary setting if the user
     *   specified a command line setting. 
     */
    set_file_safety_level(0);
    temp_file_safety_level_set_ = FALSE;
}

CHtmlPreferences::~CHtmlPreferences()
{
    /* delete the property list */
    delete proplist_;
}


/*
 *   Run the preferences dialog 
 */
void CHtmlPreferences::run_preferences_dlg(HWND hwndOwner,
                                           CHtmlSysWin_win32 *win)
{
    PROPSHEETPAGE psp[6];
    PROPSHEETHEADER psh;
    CHtmlDialogFonts *fonts_dlg;
    CHtmlDialogColor *color_dlg;
    CHtmlDialogKeys *keys_dlg;
    CHtmlDialogMore *more_dlg;
    CHtmlDialogSafety *safety_dlg;
    CHtmlDialogMem *mem_dlg;
    
    /* remember the owning window */
    win_ = win;

    /* create our dialog objects */
    fonts_dlg = new CHtmlDialogFonts(this);
    color_dlg = new CHtmlDialogColor(this);
    keys_dlg = new CHtmlDialogKeys(this);
    more_dlg = new CHtmlDialogMore(this);
    safety_dlg = new CHtmlDialogSafety(this);
    mem_dlg = new CHtmlDialogMem(this);

    /* 
     *   set up the page descriptors 
     */
    
    psp[0].dwSize = sizeof(PROPSHEETPAGE);
    psp[0].dwFlags = PSP_USETITLE;
    psp[0].hInstance = CTadsApp::get_app()->get_instance();
    psp[0].pszTemplate = MAKEINTRESOURCE(DLG_FONT);
    psp[0].pszIcon = 0;
    psp[0].pszTitle = MAKEINTRESOURCE(IDS_FONT);
    fonts_dlg->prepare_prop_page(&psp[0]);
    
    psp[1].dwSize = sizeof(PROPSHEETPAGE);
    psp[1].dwFlags = PSP_USETITLE;
    psp[1].hInstance = CTadsApp::get_app()->get_instance();
    psp[1].pszTemplate = MAKEINTRESOURCE(DLG_COLORS);
    psp[1].pszIcon = 0;
    psp[1].pszTitle = MAKEINTRESOURCE(IDS_COLORS);
    color_dlg->prepare_prop_page(&psp[1]);

    psp[2].dwSize = sizeof(PROPSHEETPAGE);
    psp[2].dwFlags = PSP_USETITLE;
    psp[2].hInstance = CTadsApp::get_app()->get_instance();
    psp[2].pszTemplate = MAKEINTRESOURCE(DLG_KEYS);
    psp[2].pszIcon = 0;
    psp[2].pszTitle = MAKEINTRESOURCE(IDS_KEYS);
    keys_dlg->prepare_prop_page(&psp[2]);

    psp[3].dwSize = sizeof(PROPSHEETPAGE);
    psp[3].dwFlags = PSP_USETITLE;
    psp[3].hInstance = CTadsApp::get_app()->get_instance();
    psp[3].pszTemplate = MAKEINTRESOURCE(DLG_MORE);
    psp[3].pszIcon = 0;
    psp[3].pszTitle = MAKEINTRESOURCE(IDS_MORE);
    more_dlg->prepare_prop_page(&psp[3]);

    psp[4].dwSize = sizeof(PROPSHEETPAGE);
    psp[4].dwFlags = PSP_USETITLE;
    psp[4].hInstance = CTadsApp::get_app()->get_instance();
    psp[4].pszTemplate = MAKEINTRESOURCE(DLG_FILE_SAFETY);
    psp[4].pszIcon = 0;
    psp[4].pszTitle = MAKEINTRESOURCE(IDS_FILE_SAFETY);
    safety_dlg->prepare_prop_page(&psp[4]);

    psp[5].dwSize = sizeof(PROPSHEETPAGE);
    psp[5].dwFlags = PSP_USETITLE;
    psp[5].hInstance = CTadsApp::get_app()->get_instance();
    psp[5].pszTemplate = MAKEINTRESOURCE(DLG_MEM);
    psp[5].pszIcon = 0;
    psp[5].pszTitle = MAKEINTRESOURCE(IDS_MEM);
    mem_dlg->prepare_prop_page(&psp[5]);

    /* set up the main dialog descriptor */
    psh.dwSize = sizeof(PROPSHEETHEADER);
    psh.dwFlags = PSH_PROPSHEETPAGE;
    psh.hwndParent = hwndOwner;
    psh.hInstance = CTadsApp::get_app()->get_instance();
    psh.pszIcon = 0;
    psh.pszCaption = (LPSTR)"Options";
    psh.nPages = sizeof(psp) / sizeof(psp[0]);
    psh.nStartPage = 0;
    psh.ppsp = (LPCPROPSHEETPAGE)&psp;
    psh.pfnCallback = 0;

    /* run the property sheet */
    PropertySheet(&psh);

    /* delete the dialogs */
    delete fonts_dlg;
    delete color_dlg;
    delete keys_dlg;
    delete more_dlg;
    delete safety_dlg;
    delete mem_dlg;
}


/*
 *   Schedule reformatting of the window 
 */
void CHtmlPreferences::schedule_reformat(int more_mode)
{
    /* 
     *   if we're only reformatting MORE mode, and the window isn't in
     *   MORE mode, we don't need to reformat 
     */
    if (more_mode && !win_->in_more_mode())
        return;
    
    /* tell the window to schedule reformatting */
    win_->schedule_reformat(TRUE, TRUE, FALSE);
}

/*
 *   Save the preferences 
 */
void CHtmlPreferences::save()
{
    int id;
    HKEY key;
    DWORD disposition;

    /* open the registry key for our preference settings */
    key = CTadsRegistry::open_key(HKEY_CURRENT_USER, reg_key_name,
                                  &disposition, TRUE);

    /* save the properties */
    for (id = (HTML_pref_id_t)0 ; id < HTML_PREF_LAST ; ++id)
        write_to_registry((HTML_pref_id_t)id, key);

    /* save the custom colors */
    CTadsRegistry::set_key_binary(key, custclr_val_name,
                                  cust_colors_, sizeof(cust_colors_));

    /* done with the registry key */
    CTadsRegistry::close_key(key);
}

/*
 *   Restore the preferences 
 */
void CHtmlPreferences::restore()
{
    int id;
    HKEY key;
    DWORD disposition;

    /* open the registry key for our preference settings */
    key = CTadsRegistry::open_key(HKEY_CURRENT_USER, reg_key_name,
                                  &disposition, TRUE);

    /* save the properties */
    for (id = 0 ; id < HTML_PREF_LAST ; ++id)
        read_from_registry((HTML_pref_id_t)id, key);

    /* load the custom colors */
    CTadsRegistry::query_key_binary(key, custclr_val_name,
                                    cust_colors_, sizeof(cust_colors_));

    /* done with the registry key */
    CTadsRegistry::close_key(key);
}

/*
 *   Write a property to the registry
 */
void CHtmlPreferences::write_to_registry(HTML_pref_id_t id, HKEY key)
{
    char buf[256];
    size_t vallen;
    CHtmlProperty *prop;
    
    /* get the property value in string format */
    prop = proplist_->get_prop(id);
    vallen = prop->gen_str_rep(buf, sizeof(buf));

    /* write it to the registry */
    CTadsRegistry::set_key_str(key, prop->get_name(), buf, vallen);
}

/*
 *   Read a property from the registry 
 */
void CHtmlPreferences::read_from_registry(HTML_pref_id_t id, HKEY key)
{
    char buf[256];
    size_t vallen;
    CHtmlProperty *prop;

    /* get the property */
    prop = proplist_->get_prop(id);

    /* if the registry value isn't set, keep the default value */
    if (!CTadsRegistry::value_exists(key, prop->get_name()))
        return;
    
    /* read the registry value */
    vallen = CTadsRegistry::query_key_str(key, prop->get_name(),
                                          buf, sizeof(buf));

    /* set the preference property item */
    prop->parse_str_rep(buf, vallen);
}

