#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/html/win32/tadsreg.cpp,v 1.1.1.1 1999/05/09 21:32:31 MJRoberts Exp $";
#endif

/* Copyright (c) 1997 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  tadsreg.cpp - registry
Function
  
Notes
  
Modified
  10/28/97 MJRoberts  - Creation
*/

#include <windows.h>
#include <stdio.h>
#include <string.h>

#ifndef TADSHTML_H
#include "tadshtml.h"
#endif
#ifndef TADSREG_H
#include "tadsreg.h"
#endif

/*
 *   open a key 
 */
HKEY CTadsRegistry::open_key(HKEY base_key, const textchar_t *key_name,
                             DWORD *disposition, int create)
{
    LONG err;
    HKEY ret_key;
    
    if (create)
    {
        err = RegCreateKeyEx(base_key, key_name, 0, 0, 0, KEY_ALL_ACCESS,
                             0, &ret_key, disposition);
    }
    else
    {
        err = RegOpenKeyEx(base_key, key_name, 0, KEY_ALL_ACCESS, &ret_key);
    }

    /* return the key if we were successful */
    return (err == ERROR_SUCCESS) ? ret_key : 0;
}

/*
 *   close a key 
 */
void CTadsRegistry::close_key(HKEY key)
{
    RegCloseKey(key);
}

/*
 *   Get a key's value 
 */
int CTadsRegistry::query_key(HKEY key, const textchar_t *valname,
                             textchar_t *buf, size_t *bufsiz)
{
    DWORD typ;
    DWORD datasiz;

    /* query the value into the buffer */
    datasiz = *bufsiz;
    if (RegQueryValueEx(key, valname, 0, &typ, (BYTE *)buf, &datasiz)
        != ERROR_SUCCESS || typ != REG_SZ)
        return 1;

    /* set the size that we read */
    *bufsiz = (size_t)datasiz;

    /* success */
    return 0;
}

/*
 *   Get a key's value as a long 
 */
long CTadsRegistry::query_key_long(HKEY key, const textchar_t *valname)
{
    char buf[128];
    size_t bufsiz = sizeof(buf);

    /* get the key text */
    if (query_key(key, valname, buf, &bufsiz))
        return 0;

    /* parse the string into a long */
    return get_atol(buf);
}

/*
 *   Get a key's value as a boolean 
 */
int CTadsRegistry::query_key_bool(HKEY key, const textchar_t *valname)
{
    char buf[128];
    size_t bufsiz = sizeof(buf);

    /* get the key text */
    if (query_key(key, valname, buf, &bufsiz))
        return 0;

    /* if it starts with 'y' or 'Y', it's true, otherwise it's false */
    return (buf[0] == 'y' || buf[0] == 'Y');
}

/*
 *   Get a key as a string 
 */
size_t CTadsRegistry::query_key_str(HKEY key, const textchar_t *valname,
                                    textchar_t *buf, size_t bufsiz)
{
    /* get the key text */
    if (query_key(key, valname, buf, &bufsiz))
        return 0;

    /* return the length, minus the null terminator */
    return bufsiz - 1;
}

/*
 *   Get a key value as binary data 
 */
size_t CTadsRegistry::query_key_binary(HKEY key, const textchar_t *valname,
                                       void *buf, size_t bufsiz)
{
    DWORD typ;
    DWORD datasiz;

    datasiz = bufsiz;
    if (RegQueryValueEx(key, valname, 0, &typ, (BYTE *)buf, &datasiz)
        != ERROR_SUCCESS || typ != REG_BINARY)
        return 0;

    /* return the length of the buffer read */
    return (size_t)datasiz;
}

void CTadsRegistry::set_key_long(HKEY key, const textchar_t *valname,
                                 long val)
{
    textchar_t buf[128];

    /* generate a string for the number and save the string */
    sprintf(buf, "%ld", val);
    RegSetValueEx(key, valname, 0, REG_SZ,
                  (BYTE *)buf, get_strlen(buf) + sizeof(textchar_t));
}

void CTadsRegistry::set_key_str(HKEY key, const textchar_t *valname,
                                const textchar_t *str, size_t len)
{
    CStringBuf buf(str, len);
    RegSetValueEx(key, valname, 0, REG_SZ,
                  (BYTE *)buf.get(), len + sizeof(textchar_t));
}

void CTadsRegistry::set_key_bool(HKEY key, const textchar_t *valname, int val)
{
    textchar_t *valstr;
    size_t vallen;

    if (val)
    {
        valstr = "Yes";
        vallen = 4 * sizeof(textchar_t);
    }
    else
    {
        valstr = "No";
        vallen = 3 * sizeof(textchar_t);
    }

    RegSetValueEx(key, valname, 0, REG_SZ, (BYTE *)valstr, vallen);
}

void CTadsRegistry::set_key_binary(HKEY key, const textchar_t *valname,
                                   void *buf, size_t bufsiz)
{
    RegSetValueEx(key, valname, 0, REG_BINARY, (BYTE *)buf, bufsiz);
}


/*
 *   Check to see if a value is set in the registry 
 */
int CTadsRegistry::value_exists(HKEY key, const textchar_t *valname)
{
    DWORD datasize;
    DWORD datatype;

    return (RegQueryValueEx(key, valname, 0, &datatype, 0, &datasize)
            == ERROR_SUCCESS);
}

