/* $Header: d:/cvsroot/tads/html/win32/w32fndlg.h,v 1.1.1.1 1999/05/09 21:32:32 MJRoberts Exp $ */

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  w32fndlg.h - TADS win32 debugger "Find" dialog
Function
  
Notes
  
Modified
  03/11/98 MJRoberts  - Creation
*/

#ifndef W32FNDLG_H
#define W32FNDLG_H

#ifndef TADSHTML_H
#include "tadshtml.h"
#endif
#ifndef TADSDLG_H
#include "tadsdlg.h"
#endif

class CTadsDialogFind: public CTadsDialog
{
public:
    CTadsDialogFind();
    ~CTadsDialogFind();

    /*
     *   Run the dialog and get the string to find.  Puts the string into
     *   the caller's buffer and returns true if we got a string, returns
     *   false if the user cancelled the dialog. 
     */
    int run_find_dlg(class CHtmlSys_dbgmain *mainwin,
                     textchar_t *buf, size_t buflen, int *exact_case,
                     int *start_at_top, int *wrap, int *direction);

    /* initialize */
    void init();

    /* handle a command */
    int do_command(WPARAM cmd, HWND ctl);

    /*
     *   Save the past strings to a configuration object/restore them 
     */
    int save_config(class CHtmlDebugConfig *config,
                    const textchar_t *varname);
    int load_config(class CHtmlDebugConfig *config,
                    const textchar_t *varname);

private:
    /*
     *   Caller's buffer for the result string.  We'll copy the result
     *   here if the user dismisses the dialog with "Find" (rather than
     *   "Cancel").  
     */
    textchar_t *buf_;
    size_t buflen_;

    /* 
     *   caller's exact case flag - we'll set this flag to true if exact
     *   case is desired, false otherwise 
     */
    int *exact_case_;

    /*
     *   caller's start-at-top flag - we'll set this flag to true if the
     *   user wants to start the search at the top of the file, false
     *   otherwise 
     */
    int *start_at_top_;

    /*
     *   caller's wrap-around flag - we'll set this flag to true if the
     *   user wants the search to wrap around at the end of the file,
     *   false otherwise 
     */
    int *wrap_around_;

    /*
     *   caller's direction-to-search flag - we'll set this to 1 for down
     *   (forwards), -1 for up (backwards) 
     */
    int *direction_;
    
    
    /* 
     *   Stored strings from past invocations.  Each time the user enters
     *   a string and accepts it, we add the string to our list; we keep
     *   only the ten most recent strings if more than that have been
     *   entered. 
     */
    textchar_t *past_strings_[10];
};

#endif /* W32FNDLG_H */

