/* $Header: d:/cvsroot/tads/html/win32/w32snd.h,v 1.1.1.1 1999/05/09 21:32:32 MJRoberts Exp $ */

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  w32snd.h - tads html win32 sound implementation
Function
  
Notes
  
Modified
  01/31/98 MJRoberts  - Creation
*/

#ifndef W32SND_H
#define W32SND_H

#ifndef TADSHTML_H
#include "tadshtml.h"
#endif
#ifndef HTMLSYS_H
#include "htmlsys.h"
#endif
#ifndef TADSSND_H
#include "tadssnd.h"
#endif


/*------------------------------------------------------------------------ */
/*
 *   MIDI Sound implementation 
 */
class CHtmlSysSoundMidi_win32: public CHtmlSysSoundMidi, public CTadsSound
{
public:
    /*
     *   CHtmlSysSoundMidi implementation 
     */
    CHtmlSysSoundMidi_win32();
    ~CHtmlSysSoundMidi_win32();

    /* load a MIDI sound */
    int load_from_midi(class CHtmlMidi *midi);

    /* play the sound */
    int play_sound(class CHtmlSysWin *win, void (*done_func)(void *, int),
                   void *done_func_ctx, int repeat, const textchar_t *url);

    /* cancel the sound */
    void cancel_sound(class CHtmlSysWin *win, int sync);

    /* suspend if necessary to make room for another sound */
    int maybe_suspend(class CHtmlSysSound *new_sound);

    /* resume playback of suspended sound */
    void resume();

private:
    /* our MIDI player object */
    class CTadsMidiFilePlayer *player_;

    /* our file pointer */
    osfildef *fp_;
};


/*------------------------------------------------------------------------ */
/*
 *   WAV Sound implementation 
 */
class CHtmlSysSoundWav_win32: public CHtmlSysSoundWav, public CTadsSound
{
public:
    /*
     *   CHtmlSysSoundWav implementation 
     */

    CHtmlSysSoundWav_win32();
    ~CHtmlSysSoundWav_win32();

    /* load a WAV sound */
    int load_from_wav(class CHtmlWav *wav);

    /* play the sound */
    int play_sound(class CHtmlSysWin *win, void (*done_func)(void *, int),
                   void *done_func_ctx, int repeat, const textchar_t *url);

    /* cancel the sound */
    void cancel_sound(class CHtmlSysWin *win, int sync);

    /* suspend if necessary to make room for another sound */
    int maybe_suspend(class CHtmlSysSound *new_sound);

    /* resume playback of suspended sound */
    void resume();

    /* 
     *   notification function to be invoked when my controlling window
     *   receives a WAV-done message 
     */
    void on_sound_done(int repeat_count);

private:
    /* our WAV player object */
    class CTadsWavFilePlayer *player_;

    /* our file object */
    HANDLE hfile_;

    /* our termination callback information */
    void (*done_func_)(void *, int);
    void *done_ctx_;

    /* 
     *   flag indicating that we're playing - the sound can only be
     *   playing in one layer at any given time; we use this flag to
     *   disallow starting the sound if it's already playing in another
     *   layer 
     */
    int playing_ : 1;
};

/* ------------------------------------------------------------------------ */
/*
 *   MPEG 2.0 audio layer 1/2/3 sound implementation 
 */
class CHtmlSysSoundMpeg_win32: public CHtmlSysSoundMpeg, public CTadsSound
{
public:
    /*
     *   CHtmlSysSoundMpeg implementation 
     */

    CHtmlSysSoundMpeg_win32();
    ~CHtmlSysSoundMpeg_win32();

    /* load from a file */
    int load_from_mpeg(class CHtmlMpeg *mpeg);

    /* play the sound */
    int play_sound(class CHtmlSysWin *win, void (*done_func)(void *, int),
                   void *done_func_ctx, int repeat, const textchar_t *url);

    /* cancel the sound */
    void cancel_sound(class CHtmlSysWin *win, int sync);

    /* suspend if necessary to make room for another sound */
    int maybe_suspend(class CHtmlSysSound *new_sound);

    /* resume playback of suspended sound */
    void resume();

    /* 
     *   notification function to be invoked when my controlling window
     *   receives a WAV-done message 
     */
    void on_sound_done(int repeat_count);

private:
    /* our WAV player object */
    class CMpegAmpW32 *player_;

    /* our file object */
    HANDLE hfile_;

    /* our termination callback information */
    void (*done_func_)(void *, int);
    void *done_ctx_;

    /* 
     *   flag indicating that we're playing - the sound can only be
     *   playing in one layer at any given time; we use this flag to
     *   disallow starting the sound if it's already playing in another
     *   layer 
     */
    int playing_ : 1;
};

#endif /* W32SND_H */

